import socket
import time
import types
import sys
import yaml
import copy
import logging
from logging import info, warning, critical, debug
import asyncio
import re

sys.path.append('C:\\Users\\Public\\repos\\xdart')

from xdart.modules.pySSRL_bServer.bServer.bl_communication import *
from xdart.modules.pySSRL_bServer.bServer.bl_command import *
from xdart.modules.pySSRL_bServer.bServer.BL_Error import *
from xdart.modules.pySSRL_bServer.bServer.PulpitObject import *
from xdart.modules.pySSRL_bServer.bServer.BL_Variables import *

class BL_Interaction():
    """This class is holds all of the beamline interaction objects and initilizes all of
    their functionality. This includes the beamline variables, rosetta object, beamline state configuration object, and
    subroutines for executing commands on the beamlines. Each beamline will have it's own LMD_beamlineInteraction()
     object."""
    
    def __init__(self, beamline_name, loop):
        """Sets up all of the objects discussed above. The only argument is the 'fun' unique name of the beamline that
        appears in the beamline config YAML files."""

        #Lets save the "fun" name
        self.beamline_name = beamline_name

        #Save a pointer to the loop for all the async functions
        self.loop = loop

        #Do the rest of the initialization in an async function - can't be done directly in __init__ since async tag is not allowed
        asyncio.ensure_future(self.setup(), loop=self.loop)



    async def setup(self):
        """Setup anything that depends on async functionality be available"""

        # Initialize the beamline communciations
        try:
            self.beamline = BL_SPEC(self)
            await self.beamline.setup()
            print("Finished beamline setup for '{}'".format(self.beamline_name))
        except:
            print("Caught exception in setup of beamline communication")
            raise RuntimeError("Caught exception in setup of communication for '{}': {}".format(self.beamline_name, sys.exc_info()[0]))

        #Get the mnemonic
        self.mnemonic = self.beamline.mnemonic


        #Lets initialize the SPEC Infoserver Interaction.
        try:
            print("interface:: '{}'".format(getattr(self.beamline, 'interface')))
            if self.beamline.interface == "SPEC Infoserver":
                print("starting beamline interface")
                debug("Setting up beamline interface for {}".format(self.beamline_name))

                self.sis = self.SPECInfoserverInteraction(self)
                print("done starting beamline interface")
            else:
                debug("Could not find correct interface for {}".format(self.beamline_name))
                raise RuntimeError("Could not find the correct interface to initialize")
        except:
            print("Caught exception starting SPEC Infoserver Interaction Class: {}".format(sys.exc_info()[0]))
            raise RuntimeError("Caught exception in setup of SPEC Infoserver Interaction for'{}': {}".format(self.beamline_name, sys.exc_info()[0]))



        # Lets refresh all of the values in the beamline registers
        debug("bl_interaction: init beamline variables")
        self.variables = BL_Variables(self)

        self.variables.manage_beamline_variables(self.variables, self)



        #await self.beamline_variables.refresh_variables()

        #create the pulpit objects. This framework uses multiple pulpits since the SIS framework actually has a few different ways of getting info
        self.command_pulpit = PulpitObject(loop=self.loop)
        self.motor_info_pulpit = PulpitObject(loop=self.loop)
        self.counter_pulpit = PulpitObject(loop=self.loop)

        #Create a lock for the communication and execute of commands for this beam line. This is necessary since we have multiple pulpits and socket communication may need to be grouped to prevent issues with our concurrancy framework from giving the wrong information between concrurrent commands (or trying to execute concurrently)
        self.beamline.comm_lock = asyncio.Lock(loop=self.loop)

    class SPECInfoserverInteraction:
        """Low level SPEC infoserver commands. These are just out of the SPEC infoserver doc"""
        def __init__(self, bi):
            debug("SPECInfoserverInteraction: init()")
            self.bi = bi
            self.beamline = bi.beamline

            #SIS has a bunch of canned text responses. Lets put them in once as a variable within the class so we don't make any mistakes when comparing values
            self.sis_text = {}
            self.sis_text['in_control'] = "client in control."
            self.sis_text['not_in_control'] = "client not in control."
            self.sis_text['control_not_available'] = "control not available."
            self.sis_text['logging on'] = "Logging is ON."
            self.sis_text['logging off'] = "Logging is OFF."

        
        async def get_version(self):
            """Run the ?ver command for spec infoserver"""
            cmd_text = "?ver"

            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)

            await cmd.execute()
            print("cmd: {}".format(cmd))
            response = cmd.response[0]
            return response


        async def get_user(self):
            """Run the ?usr command for spec infoserver"""
            cmd_text = "?usr"

            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)

            await cmd.execute()

            response = cmd.response[0]
            return response


        async def get_console_output_buffer(self, N=None, return_after=False, get_buffer_index=False):
            """Run the ?usr command for spec infoserver"""

            if N is None and get_buffer_index is False:
                cmd_text = "?con"
            elif N is None and get_buffer_index is True:
                cmd_text = "?con idx"
            elif N is not None and return_after is True:
                cmd_text = "?con {}-".format(N)
            elif N is not None and return_after is False:
                cmd_text = "?con {}".format(N)

            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator="\n")

            await cmd.execute()

            if  N is None and get_buffer_index is True:
                response = int(cmd.response[0])
            else:
                response = cmd.response


            return response


        async def is_available(self):
            """Run ?avl. Check if SPEC is available."""

            cmd_text = "?avl"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            availability = cmd.typecastResponse([bool])[0]

            return availability


        async def is_busy(self):
            """Run ?bsy. This is the opposite of ?avl. Check if spec is busy."""
            cmd_text = "?bsy"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            busy = cmd.typecastResponse([bool])[0]

            return busy


        async def get_motor_position(self, motor_name):
            """Run ?mp"""

            cmd_text = "?mp {}".format(motor_name)
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.typecastResponse([float])[0]
            return response


        async def get_motor_information(self, motor_name):
            """Run ?mi and get the motor info. Parse result and return a dict"""

            cmd_text = "?mi {}".format(motor_name)
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.typecastResponse([int])[0]

            mi_dict = self.decode_motor_status_bits(response)

            return mi_dict


        def decode_motor_status_bits(self, status):
            """Decode the 4-bit state of the motor. See SPEC infoserver doc. Returns a dictionary"""
            mi_dict = {}
            mi_dict['motor moving'] = True if status & 0x1 else False
            mi_dict['motor disabled'] = True if status & 0x2 else False
            mi_dict['low limit active'] = True if status & 0x4 else False
            mi_dict['high limit active'] = True if status & 0x8 else False

            return mi_dict


        async def get_all_motor_mnemonics(self):
            """Get all of the motor names as a list"""
            cmd_text = "?mne"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator=", ")
            await cmd.execute()
            response = cmd.response

            print(cmd)
            print(response)
            return response


        async def get_all_counter_mnemonics(self):
            """Get all of the counter names as a list. This is undocumented in the SIS documentation but I found it in Stefans code."""
            cmd_text = "?mne c"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator=", ")
            await cmd.execute()
            response = cmd.response

            return response


        async def get_all_motor_positions(self):
            """Get all motor positions"""

            cmd_text = "?mpa"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator=", ")
            await cmd.execute()

            #Could have done the listSeparator above
            response = cmd.response
            converted_response = [float(x) for x in response]

            return converted_response


        async def get_all_motor_status(self):
            """Get all motor positions"""

            cmd_text = "?mia"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator=", ")
            await cmd.execute()

            response = cmd.response

            all_motor_status =[]
            for motor_status in response:
                all_motor_status.append(self.decode_motor_status_bits(int(motor_status)))

            return all_motor_status


        async def get_all_counters(self):
            """Get all motor positions"""

            cmd_text = "?cta"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator=", ")
            await cmd.execute()

            response = cmd.response
            converted_response = [float(x) for x in response] #Is floating point the right thing? I guess....

            return converted_response


        async def get_detector_status(self):
            """Get the detector status string. Not sure what this actually is"""

            cmd_text = "?det"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, readterminator='\r\n')
            await cmd.execute()

            response = cmd.response[0]
            print(cmd)
            return response


        async def get_all_status_motors_and_counters(self):
            """This function will eventually return everything that is returned by ?all. Right now, not sure what the
            counter statuses actually mean. Returning the string right now."""
            cmd_text = "?all"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator=", ")
            await cmd.execute()

            response = cmd.response
            #converted_response = [float(x) for x in response] #Is floating point the right thing? I guess....

            return response


        async def get_current_scan_details(self):
            """Get the detector status string. Not sure what this actually is"""

            cmd_text = "?sci"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, readterminator='\r\n')
            await cmd.execute()

            response = cmd.response[0]
            print(cmd)
            return response


        async def get_current_plot_point_index(self):
            """Return the current buffer index"""
            cmd_text = "?plt idx"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.response[0]
            print(cmd)
            return response


        async def get_plot_points(self, num_pts = None):
            """Return the full buffer or the last 'num_pts' points. Default is the full buffer"""

            if num_pts is not None: #return the last num_pts points
                cmd_text = "?plt {}".format(num_pts)
            else: #Get everything
                cmd_text = "?plt all"

            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator="/")
            await cmd.execute()

            response = cmd.response
            print("len response: {}".format(len(response)))
            converted_response = []
            t0 = time.time()
            for data_series in response:
                #c_line = list(map(float, data_series.split(", ")))
                c_line = data_series.split(", ")
                converted_response.append(c_line)

            print("conversion took {}sec".format(time.time()-t0))

            return converted_response


        async def are_we_in_control(self):
            """Check if we already have control."""
            cmd_text = "?inc"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            inc = cmd.typecastResponse([bool])[0]

            return inc


        async def get_remote_control(self):
            """Get the current plot. Not sure what this actually is"""

            cmd_text = "!rqc"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.response[0]
            print(response)
            return response


        async def release_remote_control(self):
            """Get the current plot. Not sure what this actually is"""

            cmd_text = "!rlc"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.response[0]
            print(response)
            return response


        async def execute_unix_command(self, unix_cmd):
            """Get the current plot. Not sure what this actually is. untested"""

            cmd_text = "!unx {}".format(unix_cmd)
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.response[0]

            m = re.match(r"(\d+), (.*)", response, re.DOTALL) #DOTALL catches newline characters as well", response)
            if m is not None:
                cmd_success = bool(int(m.groups()[0]))
                ret_val = m.groups()[1]
                debug("execute_unix_command: success = {}, response = '{}'".format(cmd_success, ret_val))

                return {'success': cmd_success, 'response': ret_val}
            elif response == self.sis_text['not_in_control']:
                raise SISControlError(msg='execute_unix_command: not in control. Raising error')
            else:
                raise SPECCommandError(msg="execute_unix_command: unknown response: '{}'".format(response))

            print(response)
            return response


        async def abort_command(self):
            """Abort the current SPEC command - equivelent to Ctrl^C"""
            cmd_text = "!abr"
            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            #can return not in control!
            response = cmd.response[0]
            print(response)
            return response


        async def execute_command(self, spec_cmd):
            """Execute a spec command and return the uuid of the command. In the case of SIS, this just an increasing
            number."""
            cmd_text = "!cmd {}".format(spec_cmd)

            debug("execute_command: executing '{}'".format(cmd_text))

            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.response[0]
            debug("execute_command: received'{}'".format(cmd_text))

            m = re.match(r"#(\d+)", response)
            if m is not None:
                spec_cmd_uuid = int(m.groups()[0])
                debug("execute_command: command executed as #{}".format(spec_cmd_uuid))
                return spec_cmd_uuid
            elif response == self.sis_text['not_in_control']:
                raise SISControlError(msg='execute_command: not in control. Raising error')
            else:
                raise SPECCommandError(msg="execute_command: unknown response: '{}'".format(response))


        async def retrieve_result(self):
            """Get the most recent result from SIS. Sadly SIS only keeps track of the most recent result. Returns just the last one in memory. Does not block in any way."""
            cmd_text = "?res"

            debug("retrieve_result: sending")

            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.response[0]
            debug("retrieve_result: response '{}'".format(response))

            spec_result = {}
            m = re.match(r"#(\d+), (\w+), (.*)", response, re.DOTALL) #DOTALL catches newline characters as well
            if m is not None:
                matched = m.groups()
                spec_result['uuid'] = int(matched[0])
                spec_result['status ok'] = True if matched[1] == "OK" else False
                spec_result['result'] = matched[2]
            else:
                raise SPECCommandError(message="retrieve_result: Got a result we couldn't parse")

            debug('retrieve_result: parsed: {}'.format(spec_result))
            return spec_result


        async def set_sis_logging(self, set_logging_on=None):
            """This is a poorly documented feature that appears to return enteries in the SIS log. Could be useful for
            debugging or potentially getting the response of commands in case we miss calling the ?res before executing
            another command. Need to experiment. untested"""

            if set_logging_on is None:
                set_logging_on = True

            if set_logging_on is True:
                cmd_text = "!log on"
                debug('set_sis_logging: turning on')
            else:
                cmd_text = "!log off"
                debug('set_sis_logging: turning off')

            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True)
            await cmd.execute()

            response = cmd.response[0]
            debug("set_sis_logging: response '{}'".format(response))

            if response == self.sis_text['logging on'] and set_logging_on is True:
                pass
            elif response == self.sis_text['logging off'] and set_logging_on is False:
                pass
            else:
                print("response vs loggingoff:", response == self.sis_text['logging off'])
                raise SPECCommandError(message="set_sis_logging: unknown response: '{}'".format(response))

            return response


        async def get_sis_logs(self, num_entries=None):
            """A poorly documented feature that appears to return entries of SIS logging.
            Specifiy either the number of recent entries to return or 'None' to return the full log.
            I believe this is capped at 1000 entries  from Stefan's code. untested"""

            debug('get_sis_logs: starting')
            if num_entries is not None: #return last N
                cmd_text = "!log {}".format(num_entries)
            else: #return everything
                cmd_text = "!log"

            cmd = BLCommand(self.beamline, cmd_text, needsResponse=True, listSeparator="\n")
            await cmd.execute()

            response = cmd.response
            debug("get_sis_log: response '{}'".format(response))

            return response


    ###################################################################################
    # Universal Functions
    ###################################################################################


    ######################################################################################
    # Lets add some fast diagnostic functions. These will rely on the get_many_vars
    ########################################################################################



# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports
from queue import Queue
from threading import Condition
import traceback
import time

# Other imports
from xdart.utils.containers import int_1d_data, int_2d_data

# Qt imports
from pyqtgraph import Qt

# This module imports
from xdart.utils import catch_h5py_file as catch
from xdart import utils as ut

# from icecream import ic; ic.configureOutput(prefix='', includeContext=True)


class integratorThread(Qt.QtCore.QThread):
    """Thread for handling integration. Frees main gui thread from
    intensive calculations.
    
    attributes:
        arch: int, idx of arch to integrate
        lock: Condition, lock to handle access to thread attributes
        method: str, which method to call in run
        mg_1d_args, mg_2d_args: dict, arguments for multigeometry
            integration
        sphere: EwaldSphere, object that does the integration.
    
    methods:
        bai_1d_all: Calls by arch integration 1D for all arches
        bai_1d_SI: Calls by arch integration 1D for specified arch
        bai_2d_all: Calls by arch integration 2D for all arches
        bai_2d_SI: Calls by arch integration 2D for specified arch
        load: Loads data 
        mg_1d: multigeometry 1d integration
        mg_2d: multigeometry 2d integration
        mg_setup: sets up multigeometry object
        run: main thread method.
        
    signals:
        update: empty, tells parent when new data is ready.
    """
    update = Qt.QtCore.Signal(int)
    def __init__(self, sphere, arch, file_lock, parent=None):
        super().__init__(parent)
        self.sphere = sphere
        self.arch = arch
        self.file_lock = file_lock
        self.method = None
        self.lock = Condition()
        self.mg_1d_args = {}
        self.mg_2d_args = {}
    
    def run(self):
        """Calls self.method. Catches exception where method does
        not match any attributes.
        """
        with self.lock:
            method = getattr(self, self.method)
            try:
                method()
            except KeyError:
                traceback.print_exc()

    def mg_2d(self):
        """Multigeometry integrate 2d.
        """
        self.sphere.multigeometry_integrate_2d(**self.mg_2d_args)

    def mg_1d(self):
        """Multigeometry integrate 1d.
        """
        self.sphere.multigeometry_integrate_1d(**self.mg_1d_args)

    def mg_setup(self):
        """Sets up multigeometry.
        """
        self.sphere.set_multi_geo()

    def bai_2d_all(self):
        """Integrates all arches 2d. Note, does not call sphere method
        directly, handles same functions but broken up for updates
        after each image.
        """
        with self.sphere.sphere_lock:
            self.sphere.bai_2d = int_2d_data()
        for arch in self.sphere.arches:
            arch.integrate_2d(**self.sphere.bai_2d_args, global_mask=self.sphere.global_mask)
            # self.sphere.arches[arch.idx] = arch
            self.sphere._update_bai_2d(arch)
            self.update.emit(arch.idx)
        with self.file_lock:
            with catch(self.sphere.data_file, 'a') as file:
                ut.dict_to_h5(self.sphere.bai_2d_args, file, 'bai_2d_args')

    def bai_2d_SI(self):
        """Integrate the current arch, 2d
        """
        self.arch.integrate_2d(**self.sphere.bai_2d_args, global_mask=self.sphere.global_mask)

    def bai_1d_all(self):
        """Integrates all arches 1d. Note, does not call sphere method
        directly, handles same functions but broken up for updates
        after each image.
        """
        with self.sphere.sphere_lock:
            self.sphere.bai_1d = int_1d_data()
        for arch in self.sphere.arches:
            arch.integrate_1d(**self.sphere.bai_1d_args, global_mask=self.sphere.global_mask)
            # self.sphere.arches[arch.idx] = arch
            self.sphere._update_bai_1d(arch)
            self.update.emit(arch.idx)
        with self.file_lock:
            with catch(self.sphere.data_file, 'a') as file:
                ut.dict_to_h5(self.sphere.bai_1d_args, file, 'bai_1d_args')

    def bai_1d_SI(self):
        """Integrate the current arch, 1d.
        """
        self.arch.integrate_1d(**self.sphere.bai_1d_args, global_mask=self.sphere.global_mask)
    
    def load(self):
        """Load data.
        """
        self.sphere.load_from_h5()


class fileHandlerThread(Qt.QtCore.QThread):
    """Thread class for loading data. Handles locks and waiting for
    locks to be released.
    """
    sigNewFile = Qt.QtCore.Signal(str)
    sigUpdate = Qt.QtCore.Signal()
    sigTaskStarted = Qt.QtCore.Signal()
    sigTaskDone = Qt.QtCore.Signal(str)
    
    def __init__(self, sphere, arch, file_lock, parent=None):
        """
        Parameters
        ----------
        file_lock : multiprocessing.Condition
        arch : xdart.modules.ewald.EwaldArch
        sphere : xdart.modules.ewald.EwaldSphere
        """
        super().__init__(parent)
        self.sphere = sphere
        self.arch = arch
        self.file_lock = file_lock
        self.queue = Queue()
        self.fname = sphere.data_file
        self.new_fname = None
        self.lock = Condition()
        self.running = False
        
    def run(self):
        while True:
            method_name = self.queue.get()
            try:
                self.running = True
                self.sigTaskStarted.emit()
                method = getattr(self, method_name)
                method()
            except KeyError:
                traceback.print_exc()
            self.running = False
            self.sigTaskDone.emit(method_name)
    
    def set_datafile(self):
        with self.file_lock:
            self.sphere.set_datafile(
                self.fname, save_args={'compression': 'lzf'}
            )
        self.sigNewFile.emit(self.fname)
        self.sigUpdate.emit()
    
    def update_sphere(self):
        with self.file_lock:
            self.sphere.load_from_h5(replace=False, data_only=True,
                                     set_mg=False)
    
    def load_arch(self):
        with self.file_lock:
            with catch(self.sphere.data_file, 'r') as file:
                self.arch.load_from_h5(file['arches'])
        self.sigUpdate.emit()
    
    def save_data_as(self):
        if self.new_fname is not None and self.new_fname != "":
            with self.file_lock:
                with catch(self.sphere.data_file, 'r') as f1:
                    with catch(self.new_fname, 'w') as f2:
                        for key in f1:
                            f1.copy(key, f2)
                        for attr in f1.attrs:
                            f2.attrs[attr] = f1.attrs[attr]
        self.new_fname = None



# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports

# Other imports

# Qt imports
import pyqtgraph as pg
from pyqtgraph import Qt
from pyqtgraph.parametertree import Parameter

# This module imports
from xdart.utils import catch_h5py_file as catch
from xdart import utils as ut
from .integratorUI import Ui_Form
from .sphere_threads import integratorThread
from ...widgets import rangeWidget

params = [
    {'name': 'Default', 'type': 'group', 'children': [
            {'name': 'Integrate 1D', 'type': 'group', 'children': [
                {'name': 'numpoints', 'type': 'int', 'value': 10000},
                {'name': 'unit', 'type': 'list', 'values': {
                    "2" + u"\u03B8": '2th_deg', "q (A-1)": 'q_A^-1'
                    }, 'value': '2th_deg'},
                {'name': 'radial_range', 'type': 'group', 'children': [
                        {'name': 'Low', 'type': 'float', 'value': 0.0},
                        {'name': 'High', 'type': 'float', 'value': 180.0},
                        {'name': 'Auto', 'type': 'bool', 'value': False,
                         'visible': False},
                    ]
                },
                {'name': 'azimuth_range', 'type': 'group', 'children': [
                        {'name': 'Low', 'type': 'float', 'value': -180.0},
                        {'name': 'High', 'type': 'float', 'value': 180.0},
                        {'name': 'Auto', 'type': 'bool', 'value': True},
                    ]
                },
                {'name': 'monitor', 'type': 'str', 'value': 'I0'},
                {'name': 'correctSolidAngle', 'type': 'bool', 'value': True},
                {'name': 'dummy', 'type': 'float', 'value': -1.0},
                {'name': 'delta_dummy', 'type': 'float', 'value': 0.0},
                {'name': 'Apply polarization factor', 'type': 'bool', 'value': False},
                {'name': 'polarization_factor', 'type': 'float', 'value': 0, 
                    'limits': (-1, 1)},
                {'name': 'method', 'type': 'list', 'values': [
                        "numpy", "cython", "BBox", "splitpixel", "lut", "csr", 
                        "nosplit_csr", "full_csr", "lut_ocl", "csr_ocl"
                    ], 'value':'csr'},
                {'name': 'error_model', 'type': 'list', 'values': [
                        "poisson", "azimuthal"
                    ], 'value': 'poisson'},
                {'name': 'safe', 'type': 'bool', 'value': True},
                # {'name': 'block_size', 'type': 'int', 'value': 32},
                # {'name': 'profile', 'type': 'bool', 'value': False},
                ]
            },
            {'name': 'Integrate 2D', 'type': 'group', 'children': [
                {'name': 'npt_rad', 'type': 'int', 'value': 2000},
                {'name': 'npt_azim', 'type': 'int', 'value': 2000},
                {'name': 'unit', 'type': 'list', 'values': {
                    "2" + u"\u03B8": '2th_deg', "q (A-1)": 'q_A^-1'
                    }, 'value': '2th_deg'},
                {'name': 'radial_range', 'type': 'group', 'children': [
                        {'name': 'Low', 'type': 'float', 'value': 0.0},
                        {'name': 'High', 'type': 'float', 'value': 180.0},
                        {'name': 'Auto', 'type': 'bool', 'value': False, 
                        'visible': False},
                    ]
                },
                {'name': 'azimuth_range', 'type': 'group', 'children': [
                        {'name': 'Low', 'type': 'float', 'value': -180.0},
                        {'name': 'High', 'type': 'float', 'value': 180.0},
                        {'name': 'Auto', 'type': 'bool', 'value': False, 
                        'visible': False},
                    ]
                },
                {'name': 'monitor', 'type': 'str', 'value': 'None'},
                {'name': 'correctSolidAngle', 'type': 'bool', 'value': True},
                {'name': 'dummy', 'type': 'float', 'value': -1.0},
                {'name': 'delta_dummy', 'type': 'float', 'value': 0.0},
                {'name': 'Apply polarization factor', 'type': 'bool', 'value': False},
                {'name': 'polarization_factor', 'type': 'float', 'value': 0, 
                    'limits': (-1, 1)},
                {'name': 'method', 'type': 'list', 'values': [
                        "numpy", "cython", "BBox", "splitpixel", "lut", 
                        "csr", "lut_ocl", "csr_ocl"
                    ], 'value':'csr'},
                {'name': 'error_model', 'type': 'list', 'values': [
                        "poisson", "azimuthal"
                    ], 'value': 'poisson'},
                {'name': 'safe', 'type': 'bool', 'value': True}
                ]
            }
        ]

    },
]


class integratorTree(Qt.QtWidgets.QWidget):
    """Widget for controlling integration of loaded data. Presents basic
    parameters to the user in easy to control widgets, and also
    launches menus for more advanced options.
    
    attributes:
        advancedWidget1D, advancedWidget2D: advancedParameters, pop up
            windows with advanced parameters
        azimuthalRange2D, radialRange1D, radialRange2D: rangeWidget,
            widgets which control the integration ranges for 1D and 2D
            integration
        bai_1d_pars, bai_2d_pars: pyqtgraph parameters, children of
            main parameters attribute that hold parameters related to
            1D and 2D integration
        mg_1d_pars, mg_2d_pars: unused, hold paramters for multigeometry
            integration
        mg_pars: unused, holds parameters for setting up multigeometry
        ui: Ui_Form from qtdesigner
    
    methods:
        get_args: Gets requested parameters and converts them to args
            in EwaldSphere object.
        setEnabled: Enables integration and parameter modification.
        update: Grabs args from EwaldSphere object and sets all params
            to match.
    """
    def __init__(self, sphere, arch, file_lock, parent=None):
        super().__init__(parent) 
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.sphere = sphere
        self.arch = arch
        self.file_lock = file_lock
        self.parameters = Parameter.create(
            name='integrator', type='group', children=params
        )
        self.bai_1d_pars = self.parameters.child('Default', 'Integrate 1D')
        self.bai_2d_pars = self.parameters.child('Default', 'Integrate 2D')
        """
        self.mg_pars = self.parameters.child('Multi. Geometry', 
                                             'Multi Geometry Setup')
        self.mg_1d_pars = self.parameters.child('Multi. Geometry', 
                                                'Integrate 1D')
        self.mg_2d_pars = self.parameters.child('Multi. Geometry', 
                                                'Integrate 2D')
        """
        self.radialRange1D = rangeWidget("Radial", 
                                       unit=["2" + u"\u03B8" + " (deg.)", 
                                             "q (A-1)"], 
                                       range_high=180, 
                                       points_high=1e9, 
                                       parent=self,
                                       range_low=0, 
                                       defaults=[0,180,10000])
        self.azimuthalRange2D = rangeWidget("Azimuthal", 
                                          unit="(deg.)", 
                                          range_high=180, 
                                          points_high=1e9, 
                                          parent=self,
                                          range_low=-180, 
                                          defaults=[-180,180,2000])
        self.radialRange2D = rangeWidget("Radial", 
                                       unit=["2" + u"\u03B8" + " (deg.)", 
                                             "q (A-1)"], 
                                       range_high=180, 
                                       points_high=1e9, 
                                       parent=self,
                                       range_low=0, 
                                       defaults=[0,180,2000])
        self.ui.layout1D.insertWidget(1, self.radialRange1D)
        self.ui.layout2D.insertWidget(1, self.azimuthalRange2D)
        self.ui.layout2D.insertWidget(1, self.radialRange2D)
        
        self.radialRange1D.sigRangeChanged.connect(self._set_radial_range1D)
        self.radialRange1D.sigUnitChanged.connect(self._set_radial_unit1D)
        self.radialRange1D.sigPointsChanged.connect(self._set_radial_points1D)
        
        self.radialRange2D.sigRangeChanged.connect(self._set_radial_range2D)
        self.radialRange2D.sigUnitChanged.connect(self._set_radial_unit2D)
        self.radialRange2D.sigPointsChanged.connect(self._set_radial_points2D)
        
        self.azimuthalRange2D.sigRangeChanged.connect(self._set_azimuthal_range2D)
        self.azimuthalRange2D.sigPointsChanged.connect(self._set_azimuthal_points2D)
        
        self.advancedWidget1D = advancedParameters(self.bai_1d_pars, 'bai_1d')
        self.advancedWidget1D.sigUpdateArgs.connect(self.get_args)
        self.advancedWidget2D = advancedParameters(self.bai_2d_pars, 'bai_2d')
        self.advancedWidget2D.sigUpdateArgs.connect(self.get_args)
        
        self.ui.advanced1D.clicked.connect(self.advancedWidget1D.show)
        self.ui.advanced2D.clicked.connect(self.advancedWidget2D.show)
        
        self.ui.integrate1D.clicked.connect(self.bai_1d)
        self.ui.integrate2D.clicked.connect(self.bai_2d)
        
        self.integrator_thread = integratorThread(self.sphere, self.arch,
                                                  self.file_lock)
    
    def update(self):
        """Grabs args from sphere and uses _sync_ranges and
        _update_params private methods to update.
        
        args:
            sphere: EwaldSphere, object to get args from.
        """
        self._sync_ranges()
        self._update_params()
    
    def setEnabled(self, enable):
        """Overrides parent class method. Ensures appropriate child
        widgets are enabled.
        
        args:
            enable: bool, If True widgets are enabled. If False
                they are disabled.
        """
        self.advancedWidget1D.setEnabled(enable)
        self.advancedWidget2D.setEnabled(enable)
        self.radialRange1D.setEnabled(enable)
        self.radialRange2D.setEnabled(enable)
        self.azimuthalRange2D.setEnabled(enable)
        self.ui.integrate1D.setEnabled(enable)
        self.ui.integrate2D.setEnabled(enable)
        self.ui.advanced1D.setEnabled(enable)
        self.ui.advanced2D.setEnabled(enable)
        
    
    def _sync_ranges(self):
        """Syncs the range widgets. If sphere has set range arguments,
        applies those to rangeWidgets. Otherwise, adds current values to
        sphere.
        
        args:
            sphere: EwaldSphere, object to get args from.
        """
        with self.sphere.sphere_lock:
            self._sync_range(
                self.sphere.bai_1d_args, 'radial_range', 'numpoints',
                self.radialRange1D
            )
            self._sync_range(
                self.sphere.bai_2d_args, 'radial_range', 'npt_rad',
                self.radialRange2D
            )
            self._sync_range(
                self.sphere.bai_2d_args, 'azimuth_range', 'npt_azim',
                self.azimuthalRange2D
            )
    
    def _sync_range(self, args, rkey, pkey, rwidget):
        """Generic function for syncing a rangeWidget. Blocks signals
        so that internal functions don't activate while ranges are being
        updated.
        
        args:
            args: dict, arguments to check for values
            rkey: str, key for range hi and low values
            pkey: str, key for points
            rwidget: rangeWidget, widget to sync based on args
        """
        rwidget.blockSignals(True)
        try:
            self._sync_range_hilow(args, rkey, rwidget)
            self._sync_range_points(args, pkey, rwidget)
            self._sync_unit(args, rwidget)
        finally:
            rwidget.blockSignals(False)
        
    def _sync_range_points(self, args, pkey, rwidget):
        """Syncs number of points between rwidget and args.
        
        args:
            args: dict, arguments to sync
            pkey: str, key for points value in args
            rwidget: rangeWidget, widget to sync
        """
        if pkey in args:
            rwidget.ui.points.setValue(args[pkey])
        else:
            args[pkey] = rwidget.ui.points.value()

    def _sync_range_hilow(self, args, rkey, rwidget):
        """Syncs range hi and low between rwidget and args.
        
        args:
            args: dict, arguments to sync
            rkey: str, key for hi and low list value in args
            rwidget: rangeWidget, widget to sync
        """
        if rkey in args:
            if args[rkey] is None:
                args[rkey] = [rwidget.ui.low.value(),
                             rwidget.ui.high.value()]
            else:
                rwidget.ui.low.setValue(args[rkey][0])
                rwidget.ui.high.setValue(args[rkey][1])
        else:
            args[rkey] = [rwidget.ui.low.value(), rwidget.ui.high.value()]

    def _sync_unit(self, args, rwidget):
        try:
            if args["unit"] == "2th_deg":
                rwidget.ui.units.setCurrentIndex(0)
            elif args["unit"] == "q_A^-1":
                rwidget.ui.units.setCurrentIndex(1)
        except (KeyError, AttributeError):
            pass

    def _update_params(self):
        """Grabs args from sphere and syncs parameters with them.
        
        args:
            sphere: EwaldSphere, object to get args from.
        """
        with self.sphere.sphere_lock:
            self._args_to_params(self.sphere.bai_1d_args, self.bai_1d_pars)
            self._args_to_params(self.sphere.bai_2d_args, self.bai_2d_pars)
            #self._args_to_params(self.sphere.mg_args, self.mg_pars)
        
    def get_args(self, key):
        """Updates sphere with all parameters held in integrator.
        
        args:
            sphere: EwaldSphere, object to update
            key: str, which args to update.
        """
        with self.sphere.sphere_lock:
            if key == 'bai_1d':
                self._params_to_args(self.sphere.bai_1d_args, self.bai_1d_pars)
                self._sync_range(self.sphere.bai_1d_args, 'radial_range', 
                                 'numpoints', self.radialRange1D)
            elif key == 'bai_2d':
                self._params_to_args(self.sphere.bai_2d_args, self.bai_2d_pars)
                self._sync_range(self.sphere.bai_2d_args, 'radial_range',
                                 'npt_rad', self.radialRange2D)
                self._sync_range(self.sphere.bai_2d_args, 'azimuth_range',
                                 'npt_azim', self.azimuthalRange2D)
            

    def _args_to_params(self, args, tree):
        """Takes in args dictionary and sets all parameters in tree
        to match the args.
        
        args:
            args: dict, values to use for updating tree
            tree: pyqtgraph Parameter, parameters to update
        """
        with tree.treeChangeBlocker():
            for key, val in args.items():
                if 'range' in key:
                    _range = tree.child(key)
                    if val is None:
                        _range.child("Auto").setValue(True)
                    else:
                        _range.child("Low").setValue(val[0])
                        _range.child("High").setValue(val[1])
                        _range.child("Auto").setValue(False)
                elif key == 'polarization_factor':
                    if val is None:
                        tree.child('Apply polarization factor').setValue(True)
                    else:
                        tree.child('Apply polarization factor').setValue(True)
                        tree.child(key).setValue(val)
                else:
                    try:
                        child = tree.child(key)
                    except:
                        # No specific error thrown for missing child
                        child = None
                    if child is not None:
                        if val is None:
                            child.setValue('None')
                        else:
                            child.setValue(val)
    
    def _params_to_args(self, args, tree):
        """Sets all values in args to match tree.
        
        args:
            args: dict, values to be updates
            tree: pyqtgraph Parameter, parameters used to update args
        """
        for child in tree.children():
            if 'range' in child.name():
                if child.child("Auto").value():
                    args[child.name()] = None
                else:
                    args[child.name()] = [child.child("Low").value(),
                                          child.child("High").value()]
            elif child.name() == 'polarization_factor':
                pass
            elif child.name() == 'Apply polarization factor':
                if child.value():
                    args['polarization_factor'] = \
                        tree.child('polarization_factor').value()
                else:
                    args['polarization_factor'] = None
            else:
                val = child.value()
                if val == 'None':   
                    args[child.name()] = None
                else:
                    args[child.name()] = val
    
    def _set_radial_range1D(self, low, high):
        self.bai_1d_pars.child("radial_range", "Low").setValue(low)
        self.bai_1d_pars.child("radial_range", "High").setValue(high)
        self.get_args('bai_1d')
        
    def _set_radial_unit1D(self, val):
        textval = self.bai_1d_pars.child("unit").reverse[0][val]
        self.bai_1d_pars.child("unit").setValue(textval)
        self.get_args('bai_1d')

    def _set_radial_points1D(self, val):
        self.bai_1d_pars.child("numpoints").setValue(val)
        self.get_args("bai_1d")
    
    def _set_radial_range2D(self, low, high):
        self.bai_2d_pars.child("radial_range", "Low").setValue(low)
        self.bai_2d_pars.child("radial_range", "High").setValue(high)
        self.get_args('bai_2d')
        
    def _set_radial_unit2D(self, val):
        textval = self.bai_2d_pars.child("unit").reverse[0][val]
        self.bai_2d_pars.child("unit").setValue(textval)
        self.get_args('bai_2d')

    def _set_radial_points2D(self, val):
        self.bai_2d_pars.child("npt_rad").setValue(val)
        self.get_args("bai_2d")

    def _set_azimuthal_range2D(self, low, high):
        self.bai_2d_pars.child("azimuth_range", "Low").setValue(low)
        self.bai_2d_pars.child("azimuth_range", "High").setValue(high)
        self.get_args('bai_2d')
    
    def _set_azimuthal_points2D(self, val):
        self.bai_2d_pars.child("npt_azim").setValue(val)
        self.get_args("bai_2d")
    
    def bai_1d(self, q):
        """Uses the integrator_thread attribute to call bai_1d
        """
        with self.integrator_thread.lock:
            if self.ui.all1D.isChecked() or type(self.arch.idx) != int:
                self.integrator_thread.method = 'bai_1d_all'
            else:
                self.integrator_thread.method = 'bai_1d_SI'
        self.setEnabled(False)
        self.integrator_thread.start()

    def bai_2d(self, q):
        """Uses the integrator_thread attribute to call bai_2d
        """
        with self.integrator_thread.lock:
            if self.ui.all2D.isChecked() or type(self.arch.idx) != int:
                self.integrator_thread.method = 'bai_2d_all'
            else:
                self.integrator_thread.method = 'bai_2d_SI'
        self.setEnabled(False)
        self.integrator_thread.start()


class advancedParameters(Qt.QtWidgets.QWidget):
    """Pop up window for setting more advanced integration parameters.
    
    attributes: 
        name: str, name of the window
        parameter: pyqtgraph Parameter, parameters displayed
        tree: pyqtgraph ParameterTree, tree to hold parameter
        layout: QVBoxLayout, holds tree
    
    methods:
        process_change: Handles sigTreeStateChanged signal from tree
    
    signals:
        sigUpdateArgs: str, sends own name for updating args.
    """
    sigUpdateArgs = Qt.QtCore.Signal(str)
    
    def __init__(self, parameter, name, parent=None):
        super().__init__(parent)
        self.name = name
        self.parameter = parameter
        self.tree = pg.parametertree.ParameterTree()
        self.tree.addParameters(parameter)
        self.parameter.sigTreeStateChanged.connect(self.process_change)
        self.layout = Qt.QtWidgets.QVBoxLayout(self)
        self.setLayout(self.layout)
        self.layout.addWidget(self.tree)
    
    def process_change(self, tree, changes):
        self.sigUpdateArgs.emit(self.name)

# Params for multigeometry

    """
    {'name': 'Multi. Geometry', 'type': 'group', 'children': [
            {'name': 'Multi Geometry Setup', 'type': 'group', 'children': [
                {'name': 'unit', 'type': 'list', 'values': {
                    "2" + u"\u03B8": '2th_deg', "q (A-1)": 'q_A^-1'
                    }, 'value': '2th_deg'},
                {'name': 'radial_range', 'type': 'group', 'children': [
                        {'name': 'Low', 'type': 'float', 'value': 0.0},
                        {'name': 'High', 'type': 'float', 'value': 180.0},
                        {'name': 'Auto', 'type': 'bool', 'value': False, 
                         'visible': False, 'enabled': False},
                    ]
                },
                {'name': 'azimuth_range', 'type': 'group', 'children': [
                        {'name': 'Low', 'type': 'float', 'value': -180.0},
                        {'name': 'High', 'type': 'float', 'value': 180.0},
                        {'name': 'Auto', 'type': 'bool', 'value': False, 
                         'visible': False, 'enabled': False},
                    ]
                },
                {'name': 'empty', 'type': 'float', 'value': -1.0},
                {'name': 'chi_disc', 'type': 'float', 'value': 180},
            ]
            },
            {'name': 'Integrate 1D', 'type': 'group', 'children': [
                {'name': 'npt', 'type': 'int', 'value': 10000},
                {'name': 'monitor', 'type': 'str', 'value': 'None'},
                {'name': 'correctSolidAngle', 'type': 'bool', 'value': True},
                {'name': 'Apply polarization factor', 'type': 'bool', 'value': False},
                {'name': 'polarization_factor', 'type': 'float', 'value': 0, 
                    'limits': (-1, 1)}
                ]
            },
            {'name': 'Integrate 2D', 'type': 'group', 'children': [
                {'name': 'npt_rad', 'type': 'int', 'value': 2000},
                {'name': 'npt_azim', 'type': 'int', 'value': 2000},
                {'name': 'monitor', 'type': 'str', 'value': 'None'},
                {'name': 'correctSolidAngle', 'type': 'bool', 'value': True},
                {'name': 'Apply polarization factor', 'type': 'bool', 'value': False},
                {'name': 'polarization_factor', 'type': 'float', 'value': 0, 
                    'limits': (-1, 1)}
                ]
            }
        ]

    }
    """



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mainWindow.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1800, 1100)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1800, 22))
        self.menubar.setDefaultUp(True)
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuExperiments = QtWidgets.QMenu(self.menubar)
        self.menuExperiments.setObjectName("menuExperiments")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionOpen = QtWidgets.QAction(MainWindow)
        self.actionOpen.setEnabled(True)
        self.actionOpen.setObjectName("actionOpen")
        self.actionExit = QtWidgets.QAction(MainWindow)
        self.actionExit.setObjectName("actionExit")
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addAction(self.actionExit)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuExperiments.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuExperiments.setTitle(_translate("MainWindow", "Experiments"))
        self.actionOpen.setText(_translate("MainWindow", "Open"))
        self.actionExit.setText(_translate("MainWindow", "Exit"))



# -*- coding: utf-8 -*-
"""
@author: thampy
"""

# Standard library imports
import sys, os, fnmatch, time

# Other Imports
import numpy as np

from collections import OrderedDict
from joblib import Parallel, delayed
import multiprocessing as mp

# This module imports
from lmfit.models import LinearModel

import pyFAI
from silx.io.specfile import SpecFile

xdart_dir = os.getcwd().split('xdart')[0] + 'xdart'
if __name__ == '__main__':
    if xdart_dir not in sys.path:
        sys.path.append(xdart_dir)
    
from xdart.utils import get_motor_val, query, query_yes_no
from xdart.utils import get_img_data, fit_images_2D
from xdart.modules.pySSRL_bServer.bServer_funcs import specCommand, wait_until_SPECfinished, get_console_output


def get_detector_config():
    """Gets the configuration of the detector

    Returns:
        detector_name -- Pilatus100K, 300K etc.
        pixel_sz -- pixel size in meters
        ccd_shape -- number of pixels along both directions
        orientation -- vertical/horizontal
        init_dist -- calls function get_initial_distance to esimate the rough sample detector distance
    """
    
    detector_name = get_detector_name()
    detector = pyFAI.detector_factory(detector_name)

    pixel_sz = detector.pixel1
    ccd_shape = detector.max_shape

    orientation = get_detector_orientation()
    init_dist = get_initial_distance()
    
    return detector_name, pixel_sz, ccd_shape, orientation, init_dist


def check_recalibrate():
    """Check if user wants to recalibrate"""

    calibScan = query('Enter scan name to recalibrate with existing scan. Else press enter for new calibration: ')
    return calibScan
    

def check_ready():
    """Check if detector is ready for direct beam scan"""

    ready = query_yes_no('\nAre filters in? Can detector take direct beam?')

    if ready != 'yes':
        print('Cannot proceed with direct beam scan..')
        sys.exit()
    

def get_current_state():
    """Save current pd_savepath, specfilename, and scan number to
    restore the current state after callibration
    
    Returns:
        pd_savepath -- Current PD Savepath
        filename -- Current SPEC filename
        scan_n -- Current SPEC scan number
    """
    
    print('Saving current state..\n')
    
    command = "print PD_SAVEPATH ' ' DATAFILE  ' ' SCAN_N"
    specCommand(command, queue=True)
    wait_until_SPECfinished()
    
    while 1:
        try:
            out = get_console_output(idx=1)[0]
            break
        except:
            pass
        
    pd_savepath, filename, scan_n = out.split(' ')
    return pd_savepath, filename, scan_n


def restore_state(saved_state):
    """Restore saved pd_savepath, specfilename, and scan number
    
    args:
        saved_state: Tuple containing pd_savepath, specfilename and scan number
    """
    
    pd_savepath, filename, scan_n = saved_state
    
    command = f'pd savepath {pd_savepath}; newfile_f("{filename}", {scan_n})'
    specCommand(command, queue=True)
    
    print('Saved state restored\n')
    
    
def get_detector_name():
    """Get detector type from User"""
    
    detectors_dict = {'1': 'Pilatus100K',
                      '2': 'Pilatus300K',
                      '3': 'Eiger1M',
                      '' : 'Pilatus100K'}

    while 1:
        det_number = query('Choose detector (1-3) [1] Pilatus100K (2) Pilatus300K (3) Eiger1M: ')
        try:
            detector = detectors_dict[det_number]
            print(f'Selected {detector}\n')
            return detector
        except:
            print('Choice must be a number from (1 - 3)')


def get_detector_orientation():
    """Get detector orientation from User"""
    
    orientations = {'v': 'vertical',
                    'h': 'horizontal',
                    '' : 'vertical'}

    while 1:
        orientation = query('Detector orientation ( Vertical [default] / Horizontal) [v/h]: ').lower()
        if orientation in ['h', 'v', '']:
            orientation = orientations[orientation]
            print(f'Selected orientation {orientation}\n')
            return orientation
        else:
            print('Please choose (v/h) or hit enter for default [v]: ')


def get_initial_distance(default=0.7):
    """Get rough initial sample detector distance from User"""
    
    while 1:
        init_dist = query(f'Enter rough detector sample distance in meters [default {default}]:')
        if len(init_dist) == 0:
            print(f'Using default value: {default}\n')
            return default
        try:
            init_dist = float(init_dist)
            return init_dist
        except:
            print('Please enter a numeric value')


def create_remote_paths(scan_path, img_path):
    """Create folders on SPEC machine to store calibration scan and images if they don't exist"""
    
    command = f'u mkdir {scan_path}'
    try:
        specCommand(command, queue=True)
    except Exception as e:
        print(e)
        print(f"Command '{command}' not sent")
        sys.exit()    

    # Create images folder on SPEC machine if it doesn't exist
    command = f'u mkdir {img_path}'
    try:
        specCommand(command, queue=True)
    except Exception as e:
        print(e)
        print(f"Command '{command}' not sent")
        sys.exit()    


def create_SPEC_file(path):
    """Create new calibration spec file with date time stamp"""
    
    timestr = time.strftime("%Y%m%d-%H%M%S")
    calibScan = f'calib_{timestr}'

    print(f'Creating new SPEC file {calibScan} for calibration scan in {path}..')
    command = f'newfile {path}/{calibScan}'
    try:
        specCommand(command, queue=True)
    except Exception as e:
        print(e)
        print(f"Command '{command}' not sent")
        sys.exit()    
        
    return calibScan
        
        
def set_PD_savepath(img_path):
    """Change PD Savepath to img_path
    
    args:
        img_path: New PD savepath
    """

    print(f'Changing PD SavePath to {img_path}')
    command = f'pd savepath {img_path}'
    try:
        specCommand(command, queue=True)
    except Exception as e:
        print(e)
        print(f"Command '{command}' not sent")
        sys.exit()
        

def get_tth_scan_range(init_dist=0.7, pixel_sz=0.000172, ccd_shape=(195,487), orientation='vertical'):
    """Estimate the range for 2th scan for calibration

    Keyword Arguments:
        init_dist {float} -- Estimate of sample detector distance (default: {0.7})
        pixel_sz {float} -- Pixel size in meters (default: {0.000172})
        ccd_shape {tuple} -- Detector size in pixels (default: {(195,487)})
        orientation {str} -- detector orientation (vertical/horizontal) (default: {'vertical'})

    Returns:
        tuple -- 2th range, number of steps (step size is 0.02 degrees)
    """
    if orientation == 'vertical':
        ccd_shape = (487, 195)

    tth = np.round(np.rad2deg( np.arctan(pixel_sz*ccd_shape[0]/2 / init_dist) ) + 0.2, 1)
    steps = len(np.arange(-tth, tth, 0.02))

    return tth, steps


def run_direct_beam_scan(tth, steps):
    """Function to run direct beam scan to calibrate

    Arguments:
        tth {float} -- maximum 2th value to scan. Scan is performed from -tth to +tth
        steps {int} -- number of steps
    """
    command = f'ascan  tth -{tth} {tth} {steps} 1'
    print(f'Running direct beam scan [{command}]')
    try:
        specCommand(command, queue=True)
    except Exception as e:
        print(e)
        print(f"Command '{command}' not sent")
        sys.exit()
        
    # Wait till Scan is finished to continue
    print('Waiting for scan to finish..')
    wait_until_SPECfinished(polling_time=5)
    time.sleep(5)
    print('Done', '\n')


def read_SPEC_file(scan_path, calibScan):
    """Read SPEC file into an object

    Arguments:
        scan_path {path} -- path where SPEC file is stored
        calibScan {str} -- Calibration file name

    Returns:
        SpecFile object -- SpecFile Object as per SILX
    """
    spec_file = SpecFile( os.path.join(scan_path, f'{calibScan}') )
    
    return spec_file


def get_img_fnames(img_path, pdi_path, calibScan):
    """Get filenames of image files used for direct beam calibration

    Arguments:
        img_path {path} -- Directory where images are stored
        pdi_path {path} -- Directory where the associated PDI files are stored
        calibScan {string} -- Name of calibration scan

    Returns:
        tuple -- Image and PDI file names
    """

    img_fnames = sorted(fnmatch.filter(os.listdir(img_path), f'*{calibScan}*.raw'))
    pdi_fnames = [f'{img_fname}.pdi' for img_fname in img_fnames
                if os.path.exists( os.path.join(pdi_path, f'{img_fname}.pdi') )]

    if len(pdi_fnames) != len(img_fnames):
        print('PDI files for all images not present..')
        
    return img_fnames, pdi_fnames
        
           
def get_TTh_w_direct_beam(img_path, img_fnames, pdi_path, pdi_fnames):
    """Get range of TTh Values that see Direct Beam

    Arguments:
        img_path {path} -- Path for image files
        img_fnames {[str]} -- image file names
        pdi_path {path} -- Path for PDI files
        pdi_fnames {[str]} -- PDI file names

    Returns:
        tuple -- 2th values, image and PDI filenames of points that see direct beam
    """
    # Get Range of TTh Values that see Direct Beam
    img_mean_vals = [ np.mean(get_img_data( os.path.join(img_path, img_fname),
                                              return_float=True ))
                     for img_fname in img_fnames]
    median_mean_vals = np.median( np.asarray(img_mean_vals) )

    for (idx, img_fname) in enumerate(img_fnames):
        img = get_img_data( os.path.join(img_path, img_fname), return_float=True )
        if np.mean(img) > median_mean_vals/2.:
            break
    if idx > 0:
        img_fnames = img_fnames[idx:-idx]
    
    for (idx, img_fname) in enumerate(img_fnames[::-1]):
        img = get_img_data( os.path.join(img_path, img_fname), return_float=True )
        if np.mean(img) > median_mean_vals/2.:
            break
    if idx > 0:
        img_fnames = img_fnames[idx:-idx]
    
    n0 = len(img_fnames)//2
    img_fnames = img_fnames[:20] + img_fnames[n0-20 : n0+21] + img_fnames[-20:]
    
    pdi_fnames = [f'{img_fname}.pdi' for img_fname in img_fnames]
    tths = [get_motor_val( os.path.join(pdi_path, pdi_fname), 'TwoTheta' )
            for pdi_fname in pdi_fnames]
    
    return img_fnames, pdi_fnames, tths

      
def fit_all_images(img_path, img_fnames, tths, n_cores=None):
    """Perform 2D pvoigt function fits to all images to get the pixel position of the direct beam 

    Arguments:
        img_path {path} -- image file path
        img_fnames {[str]} -- image file names
        tths {[float]} -- 2th values for each image

    Keyword Arguments:
        n_cores {int} -- Number of cores to use for parallel processing (default: {None})

    Returns:
        tuple -- the direct beam pixel coordinates for each tth value and the tth values
    """
    if n_cores is None:
        n_cores = mp.cpu_count()
        
    fit_results = Parallel(n_jobs=n_cores) (
        delayed(fit_images_2D)(os.path.join(img_path, fname), tth, function='pvoigt',
                               kernel_size=3, window_size=1,
                               verbose='True', orientation='vertical', flip=False)
        for (fname, tth) in zip(img_fnames[::], tths[::]) )


    Fit_Results = OrderedDict ( { tth:fit_result for (tth, fit_result) in fit_results } )

    tths = np.asarray( list(Fit_Results.keys()) )
    xs = np.asarray( [Fit_Results[tth].params['center_x'] for tth in tths] )
    ys = np.asarray( [Fit_Results[tth].params['center_y'] for tth in tths] )
    
    return tths, xs, ys

          
def get_db_pixel(tths, xs, ys):
    """Get direct beam pixel at TTH = 0 degrees. This is done by taking a 
    total of 40 images around TTH = 0, finding the direct beam pixel for each of them
    and fitting these positions to a line to get the direct beam pixel position at TTH = 0

    Arguments:
        tths {[float]} -- array of 2th values
        xs {[float]} -- array of the x positions of the direct beam pixel
        ys {[float]} -- array of the y positions of the direct beam pixel

    Returns:
        tuple -- x0, y0 (direct beam pixel position at 2th = 0)
    """

    n0 = len(tths)//2
    s_ = np.s_[n0-20:n0+20]

    xs, ys, tths = xs[s_], ys[s_], tths[s_]

    mod_xs = LinearModel()
    params_xs = mod_xs.guess(xs, x=tths)
    fit_xs = mod_xs.fit(xs, params=params_xs, x=tths)
    fit_xs = mod_xs.fit(xs, params=fit_xs.params, x=tths)

    mod_ys = LinearModel()
    params_ys = mod_ys.guess(ys, x=tths)
    fit_ys = mod_ys.fit(ys, params=params_ys, x=tths)
    fit_ys = mod_ys.fit(ys, params=fit_ys.params, x=tths)
    
    x0 = fit_xs.eval(params=fit_xs.params, x=0.0)
    y0 = fit_ys.eval(params=fit_ys.params, x=0.0)

    return x0, y0
          
          
def get_poni_params(tths, xs, ys, x0, y0, pixel_sz=0.000172):
    """Calculate PONI file parameters from the direct beam pixel values over the 2th scan

    Arguments:
        tths {[float]} -- 2th values for the scan
        xs {[float]} -- pixel positions along X for the points of the scan
        ys {[float]} -- pixel positions along Y for the points of the scan
        x0 {float} -- pixel positions along X for direct beam at 2th=0
        y0 {float} -- pixel positions along Y for direct beam at 2th=0

    Keyword Arguments:
        pixel_sz {float} -- Pixel size (default: {0.000172})

    Returns:
        dictionary -- dictionary containing poni file parameters
    """
    xs = np.round(xs, 2)
    ys = np.round(ys, 2)
    
    n0 = len(tths)//2; tth0 = tths[n0]

    xys = []
    ds, Ds, alphas, rot3s, all_tths = [], [], [], [], []
    for ii in range(0, n0):
        idx1, idx2 = ii, -ii-1
        tth1, tth2 = tths[idx1], tths[idx2]
        
        if abs(tth1) < 2.8:
            break

        if tth1 != -tth2:
            print(f'tth1 ({tth1}) != -tth2 ({tth2})')
            continue
            
        x1, x2 = xs[idx1], xs[idx2]
        y1, y2 = ys[idx1], ys[idx2]

        xys.append(np.asarray([x1, y1, x2, y2]))

        d1 = np.sqrt( (x1 - x0)**2 + (y1 - y0)**2 ) * pixel_sz
        d2 = np.sqrt( (x2 - x0)**2 + (y2 - y0)**2 ) * pixel_sz

        ds.append(np.asarray([d1, d2]))

        tth = np.deg2rad (abs(tth1))

        D = 2 * d1 * d2 * (d1 + d2) * np.tan(tth) / (
            (d1 + d2)**2 * np.tan(tth)**2  + (d1 - d2)**2 )

        alpha = - np.arctan( 1/np.tan(tth) * (d1 - d2) / (d1 + d2) )
        rot3 = - (x1 - x2) / (y1 - y2)

        Ds.append(D)
        alphas.append(alpha)
        rot3s.append(rot3)
        all_tths.append( np.rad2deg(tth) )

    D = np.mean(Ds[:-1])
    rot2 = np.mean(alphas[:-1])
    rot3 = np.mean(rot3s[:-1])

    poni1 = y0 * pixel_sz - D*np.tan(rot2)*np.cos(rot3)
    poni2 = x0 * pixel_sz + D*np.tan(rot2)*np.sin(rot3)
    #poni1 = (485-y0) * pixel_sz + D*np.tan(rot2)*np.cos(rot3)# + 0.005

    poni_params = OrderedDict(dist=D, rot2=rot2, rot3=rot3, poni1=poni1, poni2=poni2)
    print(poni_params)
    
    return poni_params
 
 
def make_poni(poni_file, spec_file, params, detector="Pilatus100k"):
    """Create poni file from poni file parameters

    Arguments:
        poni_file {str} -- poni file name
        spec_file {SpecFile Object} -- SpecFile object that contains information form the SPEC file
        params {dictionary} -- Poni file paramter dictionary

    Keyword Arguments:
        detector {str} -- Detector type (default: {"Pilatus100k"})
    """
    energy = spec_file.motor_position_by_name(0, 'Monochrom')
    wavelength = 12398/energy * 1e-10
    print(f'Energy: {energy}, Wavelength: {wavelength}')

    poni_params = dict(
        detector = detector,
        wavelength = wavelength,
        **params)

    ai = pyFAI.azimuthalIntegrator.AzimuthalIntegrator(**poni_params)
    ai.save(poni_file)
    print(ai)
    

def run_calibration(remote_path, local_path, poni_path, public_poni_path):
    """Main function to perform calibration

    Arguments:
        remote_path {str} -- path on remote computer (SPEC)
        local_path {str} -- path on local computer (Windows)
        poni_path {str} -- path to store poni file 
        public_poni_path {str} -- path to store poni file for global access
    """

    # Remote (SPEC) Computer Paths
    remote_scan_path = f'{remote_path}/scans'
    remote_img_path  = f'{remote_path}/images'

    # Local (Windows) paths
    scan_path = os.path.join(local_path, 'scans')
    img_path  = os.path.join(local_path, 'images')
    pdi_path  = img_path

    # Make Directories if they don't exist
    print("Creating folders if they don't exist")
    os.makedirs(scan_path, exist_ok=True)
    os.makedirs(img_path, exist_ok=True)
    os.makedirs(poni_path, exist_ok=True)
    
    # Set initial detector configuration
    detector, pixel_sz, ccd_shape, orientation, init_dist = get_detector_config()
    
    # Check if user wants to recalibrate with existing scan
    calibScan = check_recalibrate()
    if calibScan:
        if calibScan in os.listdir(scan_path):
            print(f'\nRalibrating using {calibScan}')
        else:
            print(f'Calibration file {calibScan} does not exist..')
            sys.exit()
           
    else:
        # ******************* SPEC Computer Commands ********************************

        # Check if filters are in place to allow direct beam scan
        check_ready()
        
        # Save current pd_savepath, filename and scan number
        saved_state = get_current_state()

        # Create remote_paths on SPEC computer if it doesn't exist
        #create_remote_paths(remote_scan_path, remote_img_path)

        # Create SPEC file and set PD savepath
        calibScan = create_SPEC_file(remote_scan_path)
        set_PD_savepath(remote_img_path)
        
        # Get direct beam scan range and steps
        tth, steps = get_tth_scan_range(init_dist=init_dist, pixel_sz=pixel_sz,
                                        ccd_shape=ccd_shape, orientation=orientation)

        # Run direct beam scan
        run_direct_beam_scan(tth, steps)
    
        # Restore saved current pd_savepath, specfilename, and scan number
        restore_state(saved_state)
        
    # ********************** Now moving to local computer ***************************
    
    # Read SPEC File and Images from Remote (Windows) Path
    spec_file = read_SPEC_file(scan_path, calibScan)
    img_fnames, pdi_fnames = get_img_fnames(img_path, pdi_path, calibScan)
    
    # Get Range of TTh Values that see Direct Beam
    print('\nGetting all images that see direct beam')
    img_fnames, pdi_fnames, tths = get_TTh_w_direct_beam(img_path, img_fnames, 
                                                         pdi_path, pdi_fnames)
    print(f'TTh range for direct beam: [{tths[0]}, {tths[-1]}]')

    # Fit all Images to get direct beam position at all TTH values
    print('\nFitting 2D PVoigt to all images..')
    tths, xs, ys =  fit_all_images(img_path, img_fnames, tths)
    print(len(tths), len(xs))
    print('Done')

    # Get Direct Beam Pixel at TTh = 0
    print('Obtaining direct beam pixel at TTh=0')
    x0, y0 = get_db_pixel(tths, xs, ys)
    print(f'Direct beam pixel at TTh=0: [{x0:.2f}, {y0:.2f}]\n')
    
    # Obtain Poni Parameters
    print('\nCalculating PONI File parameters (dist, rot2, rot3, poni1, poni2)')
    poni_params = get_poni_params(tths, xs, ys, x0, y0, pixel_sz=pixel_sz)
    
    # Create and save PONI file
    poni_file = os.path.join(poni_path, f'{calibScan}.poni')
    print(f'\nMaking and saving {poni_file}')
    make_poni(poni_file, spec_file, poni_params, detector=detector)

    # Create and save PONI file for Public
    poni_file = os.path.join(public_poni_path, f'{calibScan}.poni')
    print(f'\nMaking and saving {poni_file} to Public Folder')
    make_poni(poni_file, spec_file, poni_params, detector=detector)


if __name__ == '__main__':
    # Paths
    remote_path = '~/data/calibration'
    local_path  = 'P:\\bl2-1\\calibration'
    poni_path   = os.path.join(local_path, 'poni_files')
    public_poni_path  = os.path.join('C:\\Users\\Public', 'poni_files')
    
    run_calibration(remote_path, local_path, poni_path, public_poni_path)



import socket
import time
import types
import sys
import yaml
import copy
import logging
from logging import info, warning, critical, debug
import asyncio
from contextlib import contextmanager
import sys, traceback


class FrameObject():
    """Inheritable class with functions common to the Control and Oracle Frames. """

    def _common_init(self, frame_pool, bi_list):
        """Takes care of creating the standard set of start/stop futures, cleaning up the frame_list in the frame pool,
         creating queues for followers, and adding the frame to the frame pool."""
        #Keep a reference to the FramePool object


        #print("here")
        self.frame_pool = frame_pool

        self.mi_list = mi_list

        #print('before cleanup')
        #house cleaning
        self.frame_pool.cleanup_frame_list()

        #print('after cleanup')

        #Create a finished flag
        self.flag_finished = False

        #Init the delete me flag to False
        self.flag_delete = False

        #Store the time of creation
        self.time_created = time.time()

        #queue for the followers. Not sure if I need this in the control frame as well.... could be useful in the cleanup?
        self.follower_queue = asyncio.Queue(loop=self.frame_pool.loop)

        #Generate the unique ID for the Frame
        self.frame_uuid = self.frame_pool.generate_frame_uuid()

        # Enable followers to attach. This should only be disabled right before cleanup of the Control and oracle
        # frames. and enabled at the end of the init routine of the control / oracle frames
        self.allow_attachement = False

        #Create a place to store all of the followers and the associated futures
        self.followers = {}

        #Create the start and stop future objects. ensure the results
        self.start_future = {}
        self.end_future = {}
        self.init_all_futures()

        #Add the start futures to the bi_list
        self.add_start_to_bi_queues()

        #Add the new frame to the frame_pool frame_list
        self.frame_pool.frame_list.append(self)

        debug('common_init: done ')


    def generate_follower_uuid(self):
        """generate unique frame IDs. This is realtively easy to do since all ID's for all motors are generated
         by a single instance of FramePool. This is an 8-digit hexadecimal number which gives us 4,294,967,296 values before
         the sequence could repeat. Since each command takes ~10ms at least, it would take roughly 500 days for something to happen....
         Actually, nothing would happen because commands that old would be taken care of by the garbage collector long long long before
         that happened. """

        follower_uuid = uuid.uuid4().hex[:8].upper()
        return follower_uuid

    def init_all_futures(self):
        """Create a start and stop future along with a 'signal' in case we need to cancel or alert the async routine
        while running"""

        # Create a signal in case we need to cancel the command or some form of external communication
        self.external_signal = asyncio.Future(loop=self.frame_pool.loop)
        asyncio.ensure_future(self.external_signal, loop=self.frame_pool.loop)

        # Create the start/end
        for bi in self.bi_list:
            #initialize future objects
            new_start_future = asyncio.Future(loop=self.frame_pool.loop)
            new_end_future = asyncio.Future(loop=self.frame_pool.loop)

            #ensure these - basically register them with the event loop?
            asyncio.ensure_future(new_start_future, loop=self.frame_pool.loop)
            asyncio.ensure_future(new_end_future, loop=self.frame_pool.loop)

            #save these in the Frame object
            self.start_future[bi.pulpit] = new_start_future
            self.end_future[bi.pulpit] = new_end_future


    def __del__(self):
        """Remember to cancel all"""
        #cancel all the pending futures from this object.
        pass

    def add_start_to_bi_queues(self):
        """Add the start futures"""

        pass



    def attach_followers(self, var_list):
        """Add each variable to the follower. We will add a 'request_uuid' and save all the futures/vars in a struct. The ability to still attach is checked. Thus, a request is specified by both a frame and request uuid"""
        #print("attaching followers:")

        if self.allow_attachement is False:
            debug("attach_followers: Returning false")
            return None


        #print("starting attachement")

        ret_dict = {}
        for follower_var in var_list:
            #Create a new future for the follower
            new_future = asyncio.Future(loop=self.frame_pool.loop)
            asyncio.ensure_future(new_future, loop=self.frame_pool.loop)

            #Add the followers to the variable
            follower_var.add_follower(self.frame_uuid, new_future)

            #record this information with the future as the key and the variaible as the value
            ret_dict[new_future] = follower_var

        #Assign a unique request ID to identify this transation. Probably overkill, but this is to allow other other following requests to attach and be identifiable (variables may be accessed multiple times in different requests)
        request_uuid = self.generate_follower_uuid()  # something the user can use to retrieve the answer if the know the frame_uuid
        self.followers[request_uuid] = ret_dict

        #print("done attaching followers")

        return request_uuid, ret_dict


    async def get_request_data(self, requestID):
        """Read the data associated with requestID."""

        try:
            request_dict = self.followers[requestID]
        except KeyError:
            return None

        data = {}
        for (future, motor_var) in request_dict.items():
            debug("get_request_data: waiting for '{}' : {}".format(motor_var.name, future))
            await future
            f_result = future.result()
            data[motor_var] = f_result.full_results


        return data

    def request_finished(self, requestID):
        """Check the futures associated w/ this request. NON BLOCKING. Returns None if requestID is invalid. Retursn False if not finished. Returns True if all futures are finished."""

        try:
            request_dict = self.followers[requestID]
        except KeyError:
            return None

        #Check all of the futures associated with the request
        for (future, motor_var) in request_dict.items():
            if future.done() is False: #If one isn't done, we return false.
                return False

        #All of them are done.
        return True

    def allows_followers(self, required_variables): #This is overloaded
        return False

    def attach_pulpits(self):
        """Add the current frame to all of the pulpit objects associated with this object. Currently, this is all of
        the motor interactor objects that this frame is set to follow. Ideally, there should be sub-objects within
        the motor interactor (in case we can have multiple concurrent things accessing a single MI)"""
        #Add the frame to the pulpit
        for bi in self.bi_list:
            bi.pulpit.add_frame(self)






    async def wait_for_pulpit(self, *args, **kwargs):
        """Wait on the futures of this object until they are set by the pulpit"""
        print("wait_for_pulpit: kwargs: {}".format(kwargs))
        try:
            msg = kwargs['msg']
        except KeyError:
            msg = ""

        print("Starting {}".format(msg), self)
        for (pulpit_obj, future) in self.start_future.items():
            print("{} checking future. Future: {}  Pulpit: {}".format(msg, future, pulpit_obj))
            if future.done() is False:
                print('  waiting on Future')
                await future
            else:
                print('  continuing')


    def step_off_pulpit(self, *args, **kwargs):
        print("step_off_pulpit: starting")
        try:
            msg = kwargs['msg']
        except KeyError:
            msg = ""

        print('setting end futures')
        #Set the end routines
        for (pulpit_obj, future) in self.end_future.items():
            #print("run_move_position setting end. Future: {}  Pulpit: {}".format(future, pulpit_obj))
            future.set_result("  finished {}".format(time.time()))
            print("  end future set ", future)

        self.flag_finished = True


    #######################################################################################
    # Decorators for frame objects to take care of the annoying details and error handling
    #######################################################################################

    def add_pulpit_control_flow(frame_function):
        """Adds control flow for the pulpit objects. This should be the first decorator around any function that
        uses the pulpits"""
        async def pulpit_control_wrapper(self, *args, **kwargs):
            debug('add_pulpit_control: starting')
            # The 'msg' keyword is used.
            try:
                msg = kwargs['msg']
            except KeyError:
                msg = None

            try:
                debug('add_pulpit_control: starting wait_for_pulpit')
                print('add_pulpit_control: kwargs: {}'.format(kwargs))
                print('add_pulpit_control: self: {}'.format(self))
                print('add_pulpit_control: msg: {}'.format(msg))
                await self.wait_for_pulpit(self, msg=msg)

                debug('add_pulpit_control: running frame_function')
                ret_val = await frame_function(self, *args, **kwargs)

            except ExitCommandGracefully as e:
                debug('Trying to exit command gracefully: {}'.format(e.msg))


                self.flag_error = True
                self.error_message = e.msg

            except:
                debug("add_pulpit_control: got an error that cannot be handled: ", sys.exc_info()[0])

                exc_type, exc_value, exc_traceback = sys.exc_info()
                exc_type, exc_value, exc_traceback = sys.exc_info()
                print("*** print_tb:")
                traceback.print_tb(exc_traceback, limit=2, file=sys.stdout)
                print("*** print_exception:")
                # exc_type below is ignored on 3.5 and later
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                                          limit=2, file=sys.stdout)
                print("*** print_exc:")
                traceback.print_exc(limit=2, file=sys.stdout)
                print("*** format_exc, first and last line:")
                formatted_lines = traceback.format_exc().splitlines()
                print(formatted_lines[0])
                print(formatted_lines[-1])
                print("*** format_exception:")
                # exc_type below is ignored on 3.5 and later
                print(repr(traceback.format_exception(exc_type, exc_value,
                                                      exc_traceback)))
                print("*** extract_tb:")
                print(repr(traceback.extract_tb(exc_traceback)))
                print("*** format_tb:")
                print(repr(traceback.format_tb(exc_traceback)))
                print("*** tb_lineno:", exc_traceback.tb_lineno)

                raise
            else:
                return ret_val
            finally:
                debug('add_pulpit_control: stepping off pulpit')
                self.step_off_pulpit(self, msg=msg)

                print('add_pulpit_control: EVERYTHING IS DONE')

        return pulpit_control_wrapper

    def repeat_until_complete(frame_function):
        """Decorator to call a function until finished. Repeat if it returns a RetryCommand error. Otherwise, raise it up."""
        async def repeat_func(self, *args, **kwargs):
            debug('repeat_func: starting')
            while True:
                try:
                    ret_val = await frame_function(self, *args, **kwargs)
                except RetryCommand:
                    #Let the loop go again
                    pass
                except:
                    raise
                else:
                    debug("repeat_until_complete: done repeating. returning '{}'".format(ret_val))
                    self.flag_finished = True
                    return ret_val

        return repeat_func



    def handle_motor_config_error(frame_function):
        async def restore_motor_config(self, *args, **kwargs):
            """Check that all the ecessary motors have a valid (non-zero) config. This will only happen in the case of power cycling."""
            debug('handle_motor_config_error: starting')
            for mi in self.mi_list:
                print("Checking motor_config of {}".format(mi.motor_name))
                #Record the current motor state before updating user_register_two
                current_persistent_state = mi.motor_config.peek_at_persistent_variables()
                current_config = await mi.motor_variables.user_register_two.read_value(None)

                #Just make sure the original value is restored. Shouldn't be necessary but lets just avoid any potential problems
                mi.motor_variables.user_register_two.value = current_persistent_state['user_register_two']

                if current_config == 0: #There should never be any configuration when a value of zero is actually allowed as the uuid
                    debug("handle errors: restoring motor {} from memory".format(mi.motor_name))
                    debug("***** If this is looping, it means there is somehow a zero saved in the persistent_state. This should never happen but if this restoring keeps looping, just delete the persistent configs - reloading from the base config (which never should have a uuid of zero) should fix all the problems")
                    await mi.motor_config.restore_persistent_state_from_array(current_persistent_state)
                    raise RetryCommand(message='Config applied. Retrying Command')

            ret_val = await frame_function(self, *args, **kwargs)
            return ret_val

        return restore_motor_config


    def handle_communication_errors(frame_function):
        """Try fixing things if an error occurs during a frame. Reconnecting or reissuing a command. Return True if no errors are caught"""
        async def wrapper_handle_comm_error(self, *args, **kwargs):
            debug('handle_communication_errors: starting')
            try:
                msg = kwargs['msg']
            except KeyError:
                msg = ""
            debug('handle_communication_errors: getting ready to start')

            try:
                debug("{}: handle_errors: starting {}".format(msg, frame_function.__name__))
                ret_val = await frame_function(self, *args, **kwargs)

            except (ConnectionResetError, asyncio.TimeoutError, MotorCommunicationError, MotorCommandError) as e:
                #This catches basically all possible connection issues. Just need to reconnect to the motor and then we can keep going
                try:
                    critical("{} : reconnecting in handle_frame_errors: {}".format(msg, e))
                    for mi in self.mi_list:
                        while True:
                            critical("{}: Reconnecting to {}".format(msg, mi.motor_name))
                            try:
                                await mi.motor.reconnect()
                            except (asyncio.TimeoutError, MotorCommunicationError, MotorCommandError):  #
                                critical("{}: Hit a timeout error in the handle_frame_errors during reconnect".format(msg))
                            else:
                                critical("{}: Reconnected".format(msg))
                                #Once we actually connect, lets kick out of the loop for this motor interactor
                                break
                except:
                    raise
                else:
                    critical("Everything reconnected")
                    raise RetryCommand(message='Motors reconnected. Retrying command')
            except:
                raise
            else:
                debug("{}: handle_errors: everything worked".format(msg))
                return ret_val

            finally:
                debug("{}: handle_errors: finished".format(msg))

        return wrapper_handle_comm_error

    def update_orphaned_followers_afterwards(frame_function):
        """The mini-oracle update routine that updates any followers associated with this frame_uuid. Never leave a follower hanging!"""
        async def mini_oracle_frame(self, *args, **kwargs):
            debug('update_orphaned_followers_afterwards: starting')
            #Call the actual functon
            ret_val = await frame_function(self, *args, **kwargs)

            debug("mini_oracle_frame: stopping attachement")

            # Stop attachement
            self.allow_attachement = False

            t0 = time.time()
            total_followers = 0
            for mi in self.mi_list:
                var_dict = mi.motor_variables.get_motor_variables()
                # print(var_dict)
                for (var_name, motor_var) in var_dict.items():  # This is a little clunky but will work
                    # print("Looking at {}: {}".format(var_name, motor_var))
                    # Technically, we could just run the update routine... but lets check so we can be sure what we are updating...
                    num_followers = motor_var.get_followers(self.frame_uuid)
                    # print(" num_followers: ", num_followers)
                    if num_followers is not None:
                        print("Updating followers of variable '{}' on motor '{}'".format(var_name, motor_var.name))
                        await motor_var.read_value(self.frame_uuid)
                        print("  updating {}: {}s".format(motor_var.name, time.time() - t0))
                        total_followers = total_followers + 1
                        # motor_var.update_followers(oracle_frame.frame_uuid)

            print("mini_oracle_frame: total followers = {}\n\n\n\n\n\n\n".format(total_followers))
            return ret_val

        return mini_oracle_frame


from collections import namedtuple
from dataclasses import dataclass, field
import copy

import numpy as np
from pyFAI import units
import h5py

from .nzarrays import nzarray1d, nzarray2d
from .. import _utils as utils





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'c:\Users\walroth\Documents\repos\xdart\xdart\experiments\ttheta_scan\gui\tthetaUI.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from pyqtgraph.Qt import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1100, 550)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.mainSplitter = QtWidgets.QSplitter(Form)
        self.mainSplitter.setOrientation(QtCore.Qt.Horizontal)
        self.mainSplitter.setObjectName("mainSplitter")
        self.leftFrame = QtWidgets.QFrame(self.mainSplitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.leftFrame.sizePolicy().hasHeightForWidth())
        self.leftFrame.setSizePolicy(sizePolicy)
        self.leftFrame.setMinimumSize(QtCore.QSize(115, 0))
        self.leftFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.leftFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.leftFrame.setLineWidth(5)
        self.leftFrame.setObjectName("leftFrame")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.leftFrame)
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.leftSplitter = QtWidgets.QSplitter(self.leftFrame)
        self.leftSplitter.setOrientation(QtCore.Qt.Vertical)
        self.leftSplitter.setObjectName("leftSplitter")
        self.hdf5Frame = QtWidgets.QFrame(self.leftSplitter)
        self.hdf5Frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.hdf5Frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.hdf5Frame.setLineWidth(3)
        self.hdf5Frame.setObjectName("hdf5Frame")
        self.metaFrame = QtWidgets.QFrame(self.leftSplitter)
        self.metaFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.metaFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.metaFrame.setLineWidth(3)
        self.metaFrame.setObjectName("metaFrame")
        self.horizontalLayout_4.addWidget(self.leftSplitter)
        self.middleFrame = QtWidgets.QFrame(self.mainSplitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.middleFrame.sizePolicy().hasHeightForWidth())
        self.middleFrame.setSizePolicy(sizePolicy)
        self.middleFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.middleFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.middleFrame.setLineWidth(5)
        self.middleFrame.setObjectName("middleFrame")
        self.rightFrame = QtWidgets.QFrame(self.mainSplitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.rightFrame.sizePolicy().hasHeightForWidth())
        self.rightFrame.setSizePolicy(sizePolicy)
        self.rightFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.rightFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.rightFrame.setLineWidth(5)
        self.rightFrame.setObjectName("rightFrame")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.rightFrame)
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.rightSplitter = QtWidgets.QSplitter(self.rightFrame)
        self.rightSplitter.setOrientation(QtCore.Qt.Vertical)
        self.rightSplitter.setObjectName("rightSplitter")
        self.integratorFrame = QtWidgets.QFrame(self.rightSplitter)
        self.integratorFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.integratorFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.integratorFrame.setLineWidth(3)
        self.integratorFrame.setObjectName("integratorFrame")
        self.wranglerFrame = QtWidgets.QFrame(self.rightSplitter)
        self.wranglerFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.wranglerFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.wranglerFrame.setLineWidth(3)
        self.wranglerFrame.setObjectName("wranglerFrame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.wranglerFrame)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.wranglerBox = QtWidgets.QComboBox(self.wranglerFrame)
        self.wranglerBox.setObjectName("wranglerBox")
        self.verticalLayout.addWidget(self.wranglerBox)
        self.wranglerStack = QtWidgets.QStackedWidget(self.wranglerFrame)
        self.wranglerStack.setObjectName("wranglerStack")
        self.verticalLayout.addWidget(self.wranglerStack)
        self.horizontalLayout_3.addWidget(self.rightSplitter)
        self.horizontalLayout.addWidget(self.mainSplitter)

        self.retranslateUi(Form)
        self.wranglerBox.currentIndexChanged['int'].connect(self.wranglerStack.setCurrentIndex)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))




import sys
import time
import logging
import watchdog.observers
from watchdog.observers import Observer
from watchdog.events import LoggingEventHandler

watchdog.observers.inotify_buffer.InotifyBuffer.delay = 0
if __name__ == "__main__":
    #logging.basicConfig(filename='log.txt',
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s - %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S')
    path = sys.argv[1] if len(sys.argv) > 1 else '.'
    event_handler = LoggingEventHandler()
    observer = Observer()
    observer.schedule(event_handler, path, recursive=False)
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


import asyncio, sys
sys.path.append('C:\\Users\\Public\\repos\\xdart')

from xdart.modules.pySSRL_bServer.bServer.BL_Error import *

class PulpitObject():
    """The pulpit object handles the first in, first out queueing of frame objects. The concept is simple enoguh - only
    one frame can be speaking at the pulpit at any given time. In the case of the Lexium MDrive Motors, each motor interactor
    has a 'pulpit' object. For a controller that controls mutiple motors, each motor would have its own pulpit object,
    meaning there would be multiple pulpits per motor interactor (e.g. the galil's)"""
    def __init__(self, loop):

        self.loop = loop

        self.frame_queue = asyncio.Queue(loop=self.loop)

        self.executing_frame = None

        #Start the frame scheduler.
        self.consumer_future = asyncio.ensure_future(self.schedule_frames(), loop=self.loop)

    def add_frame(self, new_frame):
        #print("adding frame")
        self.frame_queue.put_nowait(new_frame)
        #print("    done adding frame")

    async def schedule_frames(self):
        """Responsible for setting the result of the 'start' (for this specific object) and waiting on the 'end' future
        associated with this object. This effectively coordinates puts one frame on the pulpit at a time."""
        while True:
            print('starting wait in schedule_frames')
            self.executing_frame = await self.frame_queue.get()

            if self.executing_frame is None: #Just in case we want to stop the PulpitObject
                print("Got null frame. Stopping Pulpit")
                break

            print("pulpit {} setting 'start' future {} in frame {}".format(self, self.executing_frame.start_future, self.executing_frame))
            start_future = self.executing_frame.start_future[self]

            print("  assigned start_future= ", start_future)

            #set the result of the start future. This will start the routine associated with this frame
            print("schedule_frames: ** Set to done: ", start_future.set_result("Start!"))

            #Wait for the frame to finish before starting the next
            end_future = self.executing_frame.end_future[self]
            print("Waiting for end_future to be set to end:", end_future)

            res_end = await end_future

            print("  pulpit {} 'end' future in frame {} completed".format(self, self.executing_frame))



    def number_waiting(self):

        return self.frame_queue.qsize()




# -*- coding: utf-8 -*-
"""
@author: walroth
"""

# Standard library imports
import os
from collections import OrderedDict
import time
import copy
import traceback
import multiprocessing as mp

# Other imports
import numpy as np

# Qt imports
import pyqtgraph as pg
from pyqtgraph import Qt
from pyqtgraph.Qt import QtCore, QtWidgets
from pyqtgraph.parametertree import ParameterTree, Parameter

# This module imports
from xdart.modules.spec import LoadSpecFile, MakePONI
from xdart.utils.containers import PONI
from xdart.modules.ewald import EwaldArch, EwaldSphere
from xdart.utils import catch_h5py_file as catch
from .wrangler_widget import wranglerWidget, wranglerThread, wranglerProcess
from .liveSpecUI import Ui_Form
from ....gui_utils import NamedActionParameter
from ....widgets import commandLine
from xdart.modules.pySSRL_bServer.watcher import Watcher
from xdart.modules.pySSRL_bServer.bServer_funcs import specCommand
from xdart.utils import get_meta_from_pdi

from .spec_wrangler import DETECTOR_DICT, MaskWidget

params = [
    {'name': 'Image Directory', 'type': 'str', 'default': ''},
    NamedActionParameter(name='image_dir_browse', title= 'Browse...'),
    {'name': 'PDI Directory', 'type': 'str', 'default': ''},
    NamedActionParameter(name='pdi_dir_browse', title= 'Browse...'),
    {'name': 'File Types', 'type': 'str', 'default': "raw, pdi"},
    {'name': 'Polling Period', 'type': 'float', 'limits': [0.01, 100], 'default': 0.1},
    {'name': 'Calibration PONI File', 'type': 'str', 'default': ''},
    NamedActionParameter(name='poni_file_browse', title= 'Browse...'),
    {'name': 'Out Directory', 'type': 'str', 'default': ''},
    NamedActionParameter(name='out_dir_browse', title= 'Browse...'),
    {'name': 'Rotation Motors', 'type': 'group', 'children': [
        {'name': 'Rot1', 'type': 'str', 'default': ''},
        {'name': 'Rot2', 'type': 'str', 'default': 'tth', 'value': 'tth'},
        {'name': 'Rot3', 'type': 'str', 'default': ''}
    ]},
    {'name': 'Calibration Angles', 'type': 'group', 'children': [
        {'name': 'Rot1', 'type': 'float', 'value': 0},
        {'name': 'Rot2', 'type': 'float', 'value': 0},
        {'name': 'Rot3', 'type': 'float', 'value': 0}
    ]},
    {'name': 'Detector', 'type': 'list', 'values': [
            "Pilatus100k"
        ],
     'value':'Pilatus100k'},
    NamedActionParameter(name='set_mask', title= 'Set mask...'),

]

class liveSpecWrangler(wranglerWidget):
    """Widget for controlling spec and reading in data as it is being
    collected. 
    
    attributes:
        fname: str, file path for data storage
        file_lock: multiprocessing Condition, multiprocess safe lock
            to ensure only one process accesses data file at a time.
        ui: Qt Ui_Form, holds all gui widgets made with qtdesigner
        specCommandLine: commandLine, widget to simulate terminal line
        commands: list, set of previously entered commands
        command_queue: Queue, used to send commands to thread
        current: int, index of current command
        file_lock, mp.Condition, process safe lock for file access
        fname: str, path to data file
        keep_trying: bool, unused
        parameters: pyqtgraph Parameter, set of parameters for
            controling widget
        scan_name: str, current scan name, used to handle syncing data
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        tree: pyqtgraph ParameterTree, tree which holds parameters
        thread: liveSpecThread, thread which controls processes to
            watch for new data and to run integration
    
    methods:
        enabled: enables or disables all gui elements.
        send_command: sends command through to spec via the bServer
        set_fname: Method to safely change file name
        set_image_dir: sets the image directory
        set_out_dir: sets the output directory for new data files
        set_pdi_dir: sets the pdi directory
        set_poni_file: sets the calibration poni file
        setup: sets up the thread attribute to ensure all parameters
            are properly synced.
        stop_watching: end the watcher process.
        update_file: update the current scan name and file path 
            attributes
    
    signals:
        finished: Should be connected to thread.finished signal
        showLabel: str, text to be set as specLabel.
        sigStart: Tells tthetaWidget to start the thread and prepare
            for new data.
        sigUpdateData: int, signals a new arch has been added.
        sigUpdateFile: (str, str), sends new scan_name and file name
            to tthetaWidget.
    """
    showLabel = Qt.QtCore.Signal(str)
    def __init__(self, fname, file_lock, parent=None):
        """fname: str, path to data file. 
        file_lock: Condition, process safe lock.
        """
        super().__init__(fname, file_lock, parent)
        
        # Setup gui elements
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.ui.startButton.clicked.connect(self.sigStart.emit)
        self.ui.stopButton.clicked.connect(self.stop_watching)
        self.specCommandLine = commandLine(self)
        self.specCommandLine.send_command = self.send_command
        self.ui.commandLayout.addWidget(self.specCommandLine)
        self.buttonSend = QtWidgets.QPushButton(self)
        self.buttonSend.setText('Send')
        self.buttonSend.clicked.connect(self.send_command)
        self.ui.commandLayout.addWidget(self.buttonSend)
        self.commands = ['']
        self.current = -1
        self.keep_trying = True
        self.showLabel.connect(self.ui.specLabel.setText)

        # Setup the parameter tree
        self.tree = ParameterTree()
        self.parameters = Parameter.create(
            name='live_spec_wrangler', type='group', children=params
        )
        self.tree.setParameters(self.parameters, showTop=False)
        self.layout = Qt.QtWidgets.QVBoxLayout(self.ui.paramFrame)
        self.layout.setContentsMargins(0,0,0,0)
        self.layout.addWidget(self.tree)
        self.parameters.child('image_dir_browse').sigActivated.connect(
            self.set_image_dir
        )
        self.parameters.child('pdi_dir_browse').sigActivated.connect(
            self.set_pdi_dir
        )
        self.parameters.child('poni_file_browse').sigActivated.connect(
            self.set_poni_file
        )
        self.parameters.child('out_dir_browse').sigActivated.connect(
            self.set_out_dir
        )
        self.parameters.sigTreeStateChanged.connect(self.update)
        
        # Setup the liveSpecThread
        self.thread = liveSpecThread(
            command_queue=self.command_queue, 
            sphere_args=self.sphere_args, 
            fname=self.fname, 
            file_lock=self.file_lock,
            mp_inputs=self._get_mp_inputs(),
            img_dir=self.parameters.child('Image Directory').value(),
            pdi_dir=self.parameters.child('PDI Directory').value(),
            out_dir=self.parameters.child('Out Directory').value(),
            filetypes=self.parameters.child('File Types').value().split(),
            pollingperiod=self.parameters.child('Polling Period').value(),
            parent=self
        )
        self.thread.showLabel.connect(self.ui.specLabel.setText)
        self.thread.sigUpdateFile.connect(self.update_file)
        self.thread.finished.connect(self.finished.emit)
        self.thread.sigUpdate.connect(self.sigUpdateData.emit)
        self.setup()

        self.mask = None
        self.mask_widget = MaskWidget()
        key = self.parameters.child("Detector").value()
        data = np.zeros(DETECTOR_DICT[key]["shape"])
        data[0, 0] = 1
        self.mask_widget.set_data(data.T)
        self.mask_widget.hide()
        self.parameters.child('set_mask').sigActivated.connect(
            self.launch_mask_widget
        )

        self.mask_widget.newMask.connect(self.set_mask)

        self.setup()
    
    def setup(self):
        """Syncs all attributes of liveSpecThread with parameters
        """
        self.thread.sphere_args.update(self.sphere_args)
        self.thread.fname = self.fname
        self.thread.mp_inputs.update(self._get_mp_inputs())
        self.thread.img_dir = self.parameters.child('Image Directory').value()
        self.thread.pdi_dir = self.parameters.child('PDI Directory').value()
        self.thread.out_dir = self.parameters.child('Out Directory').value()
        self.thread.filetypes = self.parameters.child('File Types').value().split()
        self.thread.set_queues() 
        self.thread.pollingperiod = self.parameters.child('Polling Period').value()
        self.thread.mask = self.mask

        if self.parameters.child('Rotation Motors').child('Rot2').value() == '':
            self.parameters.child('Rotation Motors').child('Rot2').setValue('tth')


    def send_command(self):
        """Sends command in command line to spec, and calls
        commandLine send_command method to add command to list of
        commands.
        """
        command = self.specCommandLine.text()
        if not (command.isspace() or command == ''):
            try:
                specCommand(command, queue=True)
            except Exception as e:
                print(e)
                print(f"Command '{command}' not sent")
        
        commandLine.send_command(self.specCommandLine)
    
    def set_image_dir(self):
        """Opens file dialogue and sets the image directory
        """
        dname = Qt.QtWidgets.QFileDialog.getExistingDirectory(self)
        if dname != '':
            self.parameters.child('Image Directory').setValue(dname)
    
    def set_pdi_dir(self):
        """Opens file dialogue and sets the pdi directory
        """
        dname = Qt.QtWidgets.QFileDialog.getExistingDirectory(self)
        if dname != '':
            self.parameters.child('PDI Directory').setValue(dname)
    
    def set_out_dir(self):
        """Opens file dialogue and sets the output directory
        """
        dname = Qt.QtWidgets.QFileDialog.getExistingDirectory(self)
        if dname != '':
            self.parameters.child('Out Directory').setValue(dname)
    
    def set_poni_file(self):
        """Opens file dialogue and sets the calibration file
        """
        # fname, _ = Qt.QtWidgets.QFileDialog().getOpenFileName()
        fname, _ = Qt.QtWidgets.QFileDialog().getOpenFileName(
            filter="PONI (*.poni *.PONI)"
        )
        if fname != '':
            self.parameters.child('Calibration PONI File').setValue(fname)

    def set_mask(self, idx, mask):
        self.mask = np.arange(mask.size)[mask.ravel() == 1]
        self.thread.mask = self.mask

    def launch_mask_widget(self):
        key = self.parameters.child("Detector").value()
        data = np.zeros(DETECTOR_DICT[key]["shape"])
        data[0, 0] = 1
        if self.mask is not None:
            _mask = np.zeros_like(data)
            _mask.ravel()[self.mask] = 1
            self.mask_widget.set_data(data.T, base=_mask)
        else:
            self.mask_widget.set_data(data.T)
        self.mask_widget.show()
    
    def _get_mp_inputs(self):
        """Organizes inputs for MakePONI from parameters.
        """
        mp_inputs = OrderedDict(
            rotations = {
                "rot1": None,
                "rot2": 'tth',
                "rot3": None
            },
            calib_rotations = {
                "rot1": 0,
                "rot2": 0,
                "rot3": 0
            },
            poni_file = None,
            spec_dict = {}
        )
        rot_mot = self.parameters.child('Rotation Motors')
        for child in rot_mot:
            if child.value() == "":
                mp_inputs['rotations'][child.name().lower()] = None
            else:
                mp_inputs['rotations'][child.name().lower()] = child.value()
        
        cal_rot = self.parameters.child('Calibration Angles')
        for child in cal_rot:
            if child.value() == 0:
                pass
            else:
                mp_inputs['calib_rotations'][child.name().lower()] = child.value()
        
        mp_inputs['poni_file'] = self.parameters.child(
                                     'Calibration PONI File').value()

        return mp_inputs
    
    def update_file(self, name, fname):
        """updates the current scan name and file path attributes, emits
        them back to main widget.
        
        args:
            name: str, scan name
            fname: str, path to data file
        """
        self.scan_name = name
        self.fname = fname
        self.sigUpdateFile.emit(name, fname)

    def enabled(self, enable):
        """Sets tree and start button to enable.
        
        args:
            enable: bool, True for enabled False for disabled.
        """
        self.tree.setEnabled(enable)
        self.ui.startButton.setEnabled(enable)
    
    def stop_watching(self):
        """Sends stop command to thread.
        """
        self.command_queue.put('stop')
    

class liveSpecThread(wranglerThread):
    """Thread for controlling watcher and integrator processes. Watcher
    checks for new data, passes paths on to integrator process for
    integration.
    
    attributes:
        command_q: mp.Queue, queue to send commands to process
        file_lock: mp.Condition, process safe lock for file access
        filetypes: list, file endings to check for in watch folder
        fname: str, path to data file.
        img_dir: str, path to image directory
        mp_inputs: dict, input parameters for MakePONI
        out_dir: str, directory to output scans
        pdi_dir: str, directory for pdi files
        pollingperiod: float, how often to check folder for updates
        queues: dict, dictionary of queues for each filetype
        scan_name: str, name of current scan
        scan_number: int, number of current scan
        input_q: mp.Queue, queue for commands sent from parent
        signal_q: mp.Queue, queue for commands sent from process
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
    
    signals:
        showLabel: str, sends out text to be used in specLabel
    
    methods:
        run: Main method, called by start
        set_queues: empties out queues in self.queues and creates new
            ones
    """
    showLabel = Qt.QtCore.Signal(str)
    def __init__(self, 
            command_queue, 
            sphere_args, 
            fname, 
            file_lock,
            mp_inputs,
            img_dir,
            pdi_dir,
            out_dir,
            filetypes,
            pollingperiod,
            parent=None):
        """command_queue: mp.Queue, queue for commands sent from parent
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        fname: str, path to data file.
        file_lock: mp.Condition, process safe lock for file access
        mp_inputs: dict, input parameters for MakePONI
        img_dir: str, path to image directory
        pdi_dir: str, path to pdi directory
        out_dir: str, directory to save scans
        filetypes: list, file types to be watched in watch folder
        pollingperiod: float, how often to check for updates in
            watch folder.
        """
        super().__init__(command_queue, sphere_args, fname, file_lock, parent)
        self.sphere_args = sphere_args
        self.mp_inputs = mp_inputs
        self.img_dir = img_dir
        self.pdi_dir = pdi_dir
        self.out_dir = out_dir
        self.filetypes = filetypes
        self.pollingperiod = pollingperiod
        self.queues = {fp: mp.Queue() for fp in filetypes}
        self.mask = None
    
    def set_queues(self):
        """Empty all current file queues, recreate them based on current
        file types.
        """
        for _, q in self.queues.items():
            self._empty_q(q)
        self.queues = {fp: mp.Queue() for fp in self.filetypes}
        
    
    def run(self):
        """Initializes watcher and integrator processes, and passes
        commands from parent. Also passes updates to parent widget.
        """
        
        # Initialize watcher
        watcher = Watcher( # Process
            watchPaths=[
                self.img_dir,
                self.pdi_dir,
            ],
            filetypes=self.filetypes,
            pollingPeriod=self.pollingperiod,
            queues=self.queues,
            command_q = self.command_q,
            daemon=True
        )
        
        # Initialize integrator
        integrator = liveSpecProcess(
            command_q=self.command_q, 
            signal_q=self.signal_q, 
            sphere_args=self.sphere_args, 
            fname=self.fname, 
            file_lock=self.file_lock, 
            queues=self.queues, 
            mp_inputs=self.mp_inputs,
            pdi_dir=self.pdi_dir,
            out_dir=self.out_dir,
            global_mask = self.mask
        )
        last=False
        integrator.start()
        watcher.start()
        
        # Main loop
        while True:
            # Check for commands from main thread
            if not self.input_q.empty():
                command = self.input_q.get()
                print(command)
                if command == 'stop':
                    self.command_q.put(command)
            
            # Check for signals from integrator
            if not self.signal_q.empty():
                signal, data = self.signal_q.get()
                if signal == 'update':
                    self.sigUpdate.emit(data)
                elif signal == 'message':
                    self.showLabel.emit(data)
                elif signal == 'new_scan':
                    self.scan_name = data[0]
                    self.fname = data[1]
                    self.sigUpdateFile.emit(data[0], data[1])
                elif signal == 'TERMINATE':
                    last = True
            if last:
                break
        
        # Cleanup processes at end of loop.
        for _, q in self.queues.items():
            self._empty_q(q)
        self._empty_q(self.signal_q)
        self._empty_q(self.command_q)
        watcher.join()
        integrator.join()
    
    def _empty_q(self, q):
        """Empty out queues.
        """
        while not q.empty():
            _ = q.get()



class liveSpecProcess(wranglerProcess):
    """Process for integrating scanning area detector data, based on
    pdi files instead of spec files. Otherwise operates very similarly
    to specProcess.
    
    attributes:
        command_q: mp.Queue, queue for commands from parent thread.
        file_lock: mp.Condition, process safe lock for file access
        fname: str, path to data file
        mp_inputs: dict, input parameters for MakePONI
        out_dir: str, path to output directory
        pdi_dir: str, path to pdi directory
        queues: dict, set of queues for updates from watcher
        scan_name: str, name of current scan
        signal_q: queue to place signals back to parent thread.
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
    
    methods:
        parse_file: Determines scan name and image index from file name
        read_pdi: Reads metadata from pdi file
        read_raw: method for reading in binary .raw files and returning
            data as a numpy array.
        run: Controls flow of integration, checking for commands,
            providing updates, and catching errors.
    """
    def __init__(self, command_q, signal_q, sphere_args, fname, file_lock, 
                 queues, mp_inputs, pdi_dir, out_dir, global_mask):
        """command_q: mp.Queue, queue for commands from parent thread.
        signal_q: queue to place signals back to parent thread.
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        fname: str, path to data file
        file_lock: mp.Condition, process safe lock for file access
        mp_inputs: dict, input parameters for MakePONI
        pdi_dir: str, path to pdi directory
        out_dir: str, path to output directory
        """
        super().__init__(command_q, signal_q, sphere_args, fname, file_lock)
        self.queues = queues
        self.mp_inputs = mp_inputs
        self.scan_name = None
        self.pdi_dir = pdi_dir
        self.out_dir = out_dir
        self.mask = global_mask
    
    def _main(self):
        """Main method. Takes in file paths from queues fed by watcher,
        reads in the metadata from pdi file and image data from raw
        file. Integrates data and stores it in hdf5 file.
        """
        # TODO: This should be _main not run
        
        # Initialize MakePONI
        self.scan_name = None
        make_poni = MakePONI()
        make_poni.inputs.update(self.mp_inputs)
        
        # Main loop.
        while True:
            # Check queues for new file paths
            for key, q in self.queues.items():
                added = q.get()
                self.signal_q.put(('message', added))
                if added == 'BREAK':
                    self.signal_q.put(('TERMINATE', None))
                    return
                elif key == 'pdi':
                    pdi_file = added
                elif key == 'raw':
                    raw_file = added
                    print(added)

            # Parse file for scan name
            scan_name, i = self.parse_file(raw_file)
            
            # If new scan has started, create new sphere object
            if scan_name != self.scan_name:
                self.scan_name = scan_name
                sphere = EwaldSphere(
                    name=scan_name,
                    data_file = os.path.join(
                        self.out_dir, scan_name + ".hdf5"
                    ),
                    **self.sphere_args
                )
                sphere.global_mask = self.mask
                sphere.save_to_h5(replace=True)
                self.signal_q.put(('new_scan', 
                                   (sphere.name, sphere.data_file)))
            while True:
                # Looks for relevant data, loops until it is found or a
                # timeout occurs
                try:
                    arr = self.read_raw(raw_file)
                    
                    # Get pdi name from raw name, ensures one for one
                    raw_fname = os.path.basename(raw_file)
                    #pdi_path = os.path.dirname(pdi_file)
                    pdi_file = os.path.join(self.pdi_dir, f'{raw_fname}.pdi') 
                    
                    print(pdi_file)

                    image_meta = self.read_pdi(pdi_file)

                    make_poni.inputs['spec_dict'] = copy.deepcopy(image_meta)

                    poni = copy.deepcopy(make_poni.run())
                    break

                except (KeyError, FileNotFoundError, AttributeError, ValueError) as e:
                    # Handle exceptions related to files not being
                    # ready yet
                    print(type(e))
                    traceback.print_tb(e.__traceback__)
            
            with self.file_lock:
                # Add data to sphere  
                sphere.add_arch(
                    calculate=True, update=True, 
                    get_sd=True, set_mg=False, idx=i, map_raw=arr, 
                    poni=PONI.from_yamdict(poni), scan_info=image_meta
                )
            self.signal_q.put(('update', i))
    
    def parse_file(self, path):
        """Generate a scan name and image index from a file name.
        
        args:
            path: str, path to be parsed
        
        returns:
            scan_name: str, name of the scan
            idx: int, index of image
        """
        _, name = os.path.split(path)
        name = name.split('.')[0]
        args = name.split('_')
        scan_name = '_'.join(args[:-1])
        idx = int(args[-1])
        return scan_name, idx
    
    def read_pdi(self, pdi_file):
        """Gets the metadata from the pdi file. Uses get_meta_from_pdi, but
        returns a single dictionary rather than two.
        
        args:
            pdi_file: str, path for pdi file
        
        returns:
            image_meta: dict, dictionary with metadata
        """
        counters, motors = get_meta_from_pdi(pdi_file)
        image_meta = {}
        image_meta.update(counters)
        image_meta.update(motors)
        return image_meta
    
    def read_raw(self, file, mask=True):
        """Reads in .raw file and returns a numpy array.
        
        args:
            file: str, path to .raw file
            mask: bool, if True clips the edges of the image
        
        returns:
            map_raw: numpy array, image data.
        """
        with open(file, 'rb') as im:
            arr = np.fromstring(im.read(), dtype='int32')
            arr = arr.reshape((195, 487))
            if mask:
                for i in range(0, 10):
                    arr[:,i] = -2.0
                for i in range(477, 487):
                    arr[:,i] = -2.0
            return arr.T



# -*- coding: utf-8 -*-
"""
@author: walroth
"""
__version__ = '0.4.2'
# Top level script for running gui based program

# Standard library imports
import sys
import gc
import os
import signal

# Qt imports
from pyqtgraph.Qt import QtGui, QtWidgets

# This module imports
from xdart.gui.mainWindow import Ui_MainWindow
from xdart.gui import tabs

# Other imports
# import multiprocessing
# multiprocessing.freeze_support()
# try:
#     multiprocessing.set_start_method('spawn')
# except RuntimeError:
#     pass


def setup_data_folders(exp_list):
    """
    Creates xdart/data folder and xdart/data/tabs folder for storing
    local data. These are ignored by gitignore.

    Parameters
    ----------
    exp_list : list, set of tabs to be

    Returns
    -------
    tab_paths : dict, paths for tabs to store data
    """
    current_directory = os.path.dirname(__file__)
    data_directory = os.path.join(current_directory, "../data")
    if not os.path.isdir(data_directory):
        os.mkdir(data_directory)
    tab_paths = {}
    for e in exp_list:
        tab_directory = os.path.join(data_directory, e)
        tab_paths[e] = tab_directory
        if not os.path.isdir(tab_directory):
            os.mkdir(tab_directory)
    return tab_paths


QMainWindow = QtWidgets.QMainWindow


class Main(QMainWindow):
    def __init__(self, tab_paths):
        """

        Parameters
        ----------
        tab_paths : dict
        """
        super().__init__()
        self.tab_paths = tab_paths
        self.tabs = {}
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.ui.actionOpen.triggered.connect(self.openFile)
        self.ui.actionExit.triggered.connect(self.exit)
        self.fname = None
        self.tabwidget = QtWidgets.QTabWidget()
        self.tabwidget.setTabsClosable(True)
        self.tabwidget.tabCloseRequested.connect(self.closeExperiment)
        self.setCentralWidget(self.tabwidget)
        self.set_tabs()
        #
        self.show()
        self.resize(1600, 920)

    def exit(self):
        try:
            for i in range(self.tabwidget.count()):
                self.closeExperiment(i)
        finally:
            self.close()
            gc.collect()
            try:
                os.killpg(os.getpid(), signal.SIGTERM)
            except ProcessLookupError:
                pass
            sys.exit(1)

    def openFile(self):
        try:
            self.tabwidget.currentWidget().open_file()
        except Exception as e:
            print(e)

    def set_tabs(self):
        for e in tabs.exp_list:
            self.ui.menuExperiments.addAction(e)
        self.ui.menuExperiments.triggered.connect(self.openExperiment)

    def openExperiment(self, q):
        if q.text() == 'ttheta_scan':
            if 'ttheta_scan' not in self.tabs:
                if self.width() < 1300:
                    self.resize(1300, self.height())
                self.tabs['ttheta_scan'] = tabs.ttheta_scan.tthetaWidget(local_path=self.tab_paths['ttheta_scan'])
                self.tabwidget.addTab(self.tabs['ttheta_scan'], 'ttheta_scan')
        elif q.text() == 'static_scan':
            if 'static_scan' not in self.tabs:
                self.tabs['static_scan'] = tabs.static_scan.staticWidget(local_path=self.tab_paths['static_scan'])
                self.tabwidget.addTab(self.tabs['static_scan'], 'static_scan')

    def closeExperiment(self, q):
        _to_close = self.tabwidget.widget(q)
        name = self.tabwidget.tabText(q)
        _to_close.close()
        self.tabwidget.removeTab(q)
        del self.tabs[name]
        gc.collect()


def main():
    # multiprocessing.freeze_support()
    os.environ['PYQTGRAPH_QT_LIB'] = 'PyQt5'
    tab_paths = setup_data_folders(tabs.exp_list)
    # app = QtGui.QApplication(sys.argv)
    app = QtWidgets.QApplication(sys.argv)
    # app.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
    mw = Main(tab_paths)
    mw.show()
    app.exec_()

    # try:
    #     os.killpg(os.getpid(), signal.SIGTERM)
    # except AttributeError or ProcessLookupError:
    #     pass


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'imageWidgetUI.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(688, 480)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.imageFrame = QtWidgets.QFrame(Form)
        self.imageFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.imageFrame.setFrameShadow(QtWidgets.QFrame.Plain)
        self.imageFrame.setLineWidth(0)
        self.imageFrame.setObjectName("imageFrame")
        self.horizontalLayout.addWidget(self.imageFrame)
        self.toolFrame = QtWidgets.QFrame(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toolFrame.sizePolicy().hasHeightForWidth())
        self.toolFrame.setSizePolicy(sizePolicy)
        self.toolFrame.setMinimumSize(QtCore.QSize(120, 0))
        self.toolFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.toolFrame.setFrameShadow(QtWidgets.QFrame.Plain)
        self.toolFrame.setLineWidth(0)
        self.toolFrame.setObjectName("toolFrame")
        self.toolLayout = QtWidgets.QGridLayout(self.toolFrame)
        self.toolLayout.setContentsMargins(1, 1, 1, 1)
        self.toolLayout.setObjectName("toolLayout")
        self.cmapBox = QtWidgets.QComboBox(self.toolFrame)
        self.cmapBox.setObjectName("cmapBox")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.cmapBox.addItem("")
        self.toolLayout.addWidget(self.cmapBox, 0, 1, 1, 1)
        self.logButton = QtWidgets.QToolButton(self.toolFrame)
        self.logButton.setMinimumSize(QtCore.QSize(30, 0))
        self.logButton.setMaximumSize(QtCore.QSize(30, 16777215))
        self.logButton.setCheckable(True)
        self.logButton.setObjectName("logButton")
        self.toolLayout.addWidget(self.logButton, 0, 0, 1, 1)
        self.horizontalLayout.addWidget(self.toolFrame)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.cmapBox.setItemText(0, _translate("Form", "grey"))
        self.cmapBox.setItemText(1, _translate("Form", "viridis"))
        self.cmapBox.setItemText(2, _translate("Form", "plasma"))
        self.cmapBox.setItemText(3, _translate("Form", "inferno"))
        self.cmapBox.setItemText(4, _translate("Form", "magma"))
        self.cmapBox.setItemText(5, _translate("Form", "spectrum"))
        self.cmapBox.setItemText(6, _translate("Form", "thermal"))
        self.cmapBox.setItemText(7, _translate("Form", "flame"))
        self.cmapBox.setItemText(8, _translate("Form", "yellowy"))
        self.cmapBox.setItemText(9, _translate("Form", "bipolar"))
        self.cmapBox.setItemText(10, _translate("Form", "cyclic"))
        self.cmapBox.setItemText(11, _translate("Form", "greyclip"))
        self.logButton.setText(_translate("Form", "Log"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())